###########################################################################
# Copyright (c) 2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import sys
from enum import unique, Enum
from typing import Iterator, List
from collections import defaultdict

from ...error_codes import IsoQuantExitCode
from ..common import load_barcodes

logger = logging.getLogger('IsoQuant')


@unique
class ElementType(Enum):
    PolyT = 1
    cDNA = 2
    CONST = 10
    VAR_ANY = 20
    VAR_LIST = 21
    VAR_FILE = 22
    VAR_ANY_SEPARATOR = 31
    VAR_ANY_NON_T_SEPARATOR = 32

    def needs_correction(self):
        return self in (ElementType.VAR_FILE, ElementType.VAR_LIST)

    def needs_sequence_extraction(self):
        return self == ElementType.VAR_ANY

    def is_base_separator(self):
        return self in (ElementType.VAR_ANY_SEPARATOR, ElementType.VAR_ANY_NON_T_SEPARATOR)

    def is_variable(self):
        return self in (ElementType.VAR_ANY, ElementType.VAR_LIST, ElementType.VAR_FILE)

    def is_constant(self):
        return self == ElementType.CONST

    def needs_only_coordinates(self):
        """Return True if element only needs start/end coordinates (no sequence extraction)."""
        return self.is_constant()


class MoleculeElement:
    def __init__(self, element_name: str, element_type: ElementType, element_value_1 = None, element_value_2 = None):
        self.element_name = element_name
        self.element_type = element_type

        if self.element_type in [ElementType.PolyT, ElementType.cDNA]:
            self.element_value = None
            self.element_length = -1
        elif self.element_type == ElementType.CONST:
            self.element_value = element_value_1
            self.element_length = len(self.element_value)
        elif self.element_type == ElementType.VAR_FILE:
            self.element_value = load_barcodes(element_value_1, needs_iterator=False)
            if element_value_2 is None:
                self.element_length = len(self.element_value[0])
            else:
                try:
                    self.element_length = int(element_value_2)
                except ValueError:
                    logger.critical("Incorrectly specified length for element %s: %s" % (self.element_name, element_value_2))
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
        elif self.element_type == ElementType.VAR_LIST:
            self.element_value = element_value_1.split(',')
            if element_value_2 is None:
                self.element_length = len(self.element_value[0])
            else:
                try:
                    self.element_length = int(element_value_2)
                except ValueError:
                    logger.critical(
                        "Incorrectly specified length for element %s: %s" % (self.element_name, element_value_2))
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
        elif self.element_type in (ElementType.VAR_ANY, ElementType.VAR_ANY_SEPARATOR, ElementType.VAR_ANY_NON_T_SEPARATOR):
            self.element_value = None
            try:
                self.element_length = int(element_value_1)
            except ValueError:
                logger.critical("Incorrectly specified length for element %s: %s" % (self.element_name, element_value_1))
                sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
        else:
            logger.critical("Wrong element type %s" % element_type)
            sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)


class MoleculeStructure:
    """
    Defines the structure of a molecule for barcode/UMI extraction.

    Parses molecule definition files (MDF) and identifies which elements
    are barcodes and UMIs based on naming convention:
    - Elements with prefix "barcode" (case-insensitive) are barcodes
    - Elements with prefix "umi" (case-insensitive) are UMIs
    """

    CONCAT_DELIM = "|"
    DUPL_DELIM = "/"

    def __init__(self, str_iterator):
        self.ordered_elements: List[MoleculeElement] = []
        self.elements_to_concatenate = {}
        self.concatenated_elements_counts = {}
        self.duplicated_elements = {}
        self.duplicated_elements_counts = {}
        # Barcode/UMI element names identified by naming convention
        self._barcode_elements: List[str] = []
        self._umi_elements: List[str] = []

        l = next(str_iterator)
        elements = list(map(lambda x: x.strip(), l.strip().split(':')))
        element_properties = {}
        for l in str_iterator:
            v = l.strip().split()
            if not v:
                continue
            element_name = v[0]
            if element_name in element_properties:
                logger.critical("Duplicated element name %s" % element_name)
                sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)

            if len(v) == 3:
                element_properties[element_name] = (v[1], v[2], None)
            elif len(v) == 4:
                element_properties[element_name] = (v[1], v[2], v[3])
            else:
                logger.critical("Incorrect number of properties in line %s" % l.strip())
                sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)

        element_name: str
        for element_name in elements:
            if element_name not in element_properties:
                if element_name not in (ElementType.cDNA.name, ElementType.PolyT.name):
                    logger.critical("Molecule element %s was not described in the format file" % element_name)
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
                element_type = ElementType[element_name]
                self.ordered_elements.append(MoleculeElement(element_name, element_type))
            else:
                element_type, element_val1, element_val2 = element_properties[element_name]
                if element_type not in ElementType.__dict__:
                    logger.critical("Molecule element type %s is not among the possible types" % element_type)
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
                element_type = ElementType[element_type]
                self.ordered_elements.append(MoleculeElement(element_name, element_type, element_val1, element_val2))

            if self.CONCAT_DELIM in element_name:
                if not (element_type.needs_sequence_extraction() or element_type.needs_correction()):
                    logger.critical("Concatenated elements must be variable (fix element %s)" % element_name)
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
                v = element_name.split(self.CONCAT_DELIM)
                if len(v) != 2:
                    logger.critical("Incorrect concatenated element %s" % element_name)
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
                try:
                    self.elements_to_concatenate[element_name] = (v[0], int(v[1]))
                except ValueError:
                    logger.critical("Incorrectly specified index for concatenated element %s: %s" % (element_name, v[1]))
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)

            elif self.DUPL_DELIM in element_name:
                if not (element_type.needs_sequence_extraction() or element_type.needs_correction()):
                    logger.critical("Concatenated elements must be variable (fix element %s)" % element_name)
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
                v = element_name.split(self.DUPL_DELIM)
                if len(v) != 2:
                    logger.critical("Incorrect duplicated element %s" % element_name)
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
                try:
                    self.duplicated_elements[element_name] = (v[0], int(v[1]))
                except ValueError:
                    logger.critical("Incorrectly specified index for duplicated element %s: %s" % (element_name, v[1]))
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)

        self._check_linked_elements(self.elements_to_concatenate)
        self.concatenated_elements_counts = self._check_indices(self.elements_to_concatenate)
        self._check_linked_elements(self.duplicated_elements)
        self.duplicated_elements_counts = self._check_indices(self.duplicated_elements)
        self._identify_barcode_umi_elements()

    def _check_linked_elements(self, element_dict):
        # check that duplicated or concatenated elements have the same values (e.g. barcode whitelist)
        # base element name -> List[MoleculeElement]
        linked_elements = defaultdict(list)
        for el in self.ordered_elements:
            if el.element_name in element_dict:
                base_element_name = element_dict[el.element_name][0]
                linked_elements[base_element_name].append(el)

        for element_list in linked_elements.values():
            if len(set(el.element_type for el in element_list)) != 1:
                logger.critical("Linked elements must have the same type (fix elements: %s)" % ", ".join(el.element_name for el in element_list))
                sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)
            first_value = element_list[0].element_value
            for el in element_list[1:]:
                if el.element_value != first_value:
                    logger.critical("Linked elements must have the same values (fix elements: %s)" % ", ".join(e.element_name for e in element_list))
                    sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)

    def _check_indices(self, element_dict):
        # check for consecutive indices
        index_dict = defaultdict(list)
        for el in element_dict:
            base_element_name = element_dict[el][0]
            index_dict[base_element_name].append(element_dict[el][1])

        for el in index_dict:
            if sorted(index_dict[el]) != list(range(1, len(index_dict[el]) + 1)):
                logger.critical("Concatenated elements must be consecutive from 1 to %d (fix element %s, indices: %s)" % (len(index_dict[el]), el, str(index_dict[el])))
                sys.exit(IsoQuantExitCode.INVALID_FILE_FORMAT)

        return {el: len(val) for el, val in index_dict.items()}

    def _identify_barcode_umi_elements(self) -> None:
        """
        Identify barcode/UMI elements by naming convention.

        Elements with prefix "barcode" (case-insensitive) are barcodes.
        Elements with prefix "umi" (case-insensitive) are UMIs.

        For concatenated/duplicated elements (e.g., barcode|1, barcode|2),
        the base name (e.g., "barcode") is used and deduplicated.
        """
        self._barcode_elements = []
        self._umi_elements = []
        seen = set()
        for el in self.ordered_elements:
            # For concatenated/duplicated elements, use the base name
            if el.element_name in self.elements_to_concatenate:
                name = self.elements_to_concatenate[el.element_name][0]
            elif el.element_name in self.duplicated_elements:
                name = self.duplicated_elements[el.element_name][0]
            else:
                name = el.element_name

            if name in seen:
                continue
            seen.add(name)

            name_lower = name.lower()
            if name_lower.startswith("barcode"):
                self._barcode_elements.append(name)
            elif name_lower.startswith("umi"):
                self._umi_elements.append(name)

    @property
    def barcode_elements(self) -> List[str]:
        """Get list of barcode element names (identified by 'barcode' prefix)."""
        return self._barcode_elements

    @property
    def umi_elements(self) -> List[str]:
        """Get list of UMI element names (identified by 'umi' prefix)."""
        return self._umi_elements

    @classmethod
    def from_element_list(cls, element_list):
        """Create MoleculeStructure from a list of MoleculeElement objects."""
        ms = cls.__new__(cls)
        ms.ordered_elements = element_list
        ms.elements_to_concatenate = {}
        ms.duplicated_elements = {}
        ms._barcode_elements = []
        ms._umi_elements = []
        ms._identify_barcode_umi_elements()
        return ms

    def __iter__(self) -> Iterator[MoleculeElement]:
        for e in self.ordered_elements:
            yield e




#!/usr/bin/env python3
#
############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Standalone barcode detection tool for single-cell and spatial transcriptomics.

This is a CLI wrapper around src.barcode_calling.detect_barcodes.
"""

import argparse
import logging
import os
import sys
from traceback import print_exc

from src.modes import IsoQuantMode
from src.error_codes import IsoQuantExitCode
from src.barcode_calling.detect_barcodes import (
    process_single_thread,
    process_in_parallel,
    BARCODE_CALLING_MODES,
)

logger = logging.getLogger('IsoQuant')


def set_logger(logger_instance, args):
    logger_instance.setLevel(logging.INFO)
    if args.debug:
        logger_instance.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    if args.debug:
        ch.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger_instance.addHandler(ch)


def parse_args(sys_argv):
    def add_hidden_option(*args, **kwargs):  # show command only with --full-help
        kwargs['help'] = argparse.SUPPRESS
        parser.add_argument(*args, **kwargs)

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--output", "-o", type=str, help="output prefix name", required=True)
    parser.add_argument("--barcodes", "-b", nargs='+', type=str, help="barcode whitelist(s)", required=False)
    parser.add_argument("--mode", type=str, help="mode to be used", choices=[x.name for x in BARCODE_CALLING_MODES.keys()])
    parser.add_argument("--molecule", type=str, help="MDF files with molecule description (for custom_sc mode only)")

    parser.add_argument("--input", "-i", nargs='+', type=str, help="input reads in [gzipped] FASTA, FASTQ, BAM, SAM",
                        required=True)
    parser.add_argument("--threads", "-t", type=int, help="threads to use (16)", default=16)
    parser.add_argument("--tmp_dir", type=str, help="folder for temporary files")
    add_hidden_option('--debug', action='store_true', default=False, help='Debug log output.')

    args = parser.parse_args(sys_argv)
    args.mode = IsoQuantMode[args.mode]
    args.out_fasta = None
    args.output_tsv = None
    return args


def check_args(args):
    """Set up output file lists based on input files."""
    # args.input is always a list (nargs='+')
    num_files = len(args.input)

    if args.output_tsv is None:
        if num_files == 1:
            args.output_tsv = [args.output + ".barcoded_reads.tsv"]
        else:
            args.output_tsv = [args.output + "_%d.barcoded_reads.tsv" % i for i in range(num_files)]

    if args.out_fasta is None and args.mode.produces_new_fasta():
        if num_files == 1:
            args.out_fasta = [args.output + ".split_reads.fasta"]
        else:
            args.out_fasta = [args.output + "_%d.split_reads.fasta" % i for i in range(num_files)]


def main(sys_argv):
    args = parse_args(sys_argv)
    set_logger(logger, args)
    check_args(args)

    out_dir = os.path.dirname(args.output)
    if out_dir and not os.path.exists(out_dir):
        os.makedirs(out_dir, exist_ok=True)

    if args.threads == 1 or args.mode.enforces_single_thread():
        process_single_thread(args)
    else:
        process_in_parallel(args)


def main_entry():
    """Entry point for console_scripts (pip install)."""
    try:
        main(sys.argv[1:])
    except SystemExit:
        raise
    except:
        print_exc()
        sys.exit(IsoQuantExitCode.UNCAUGHT_EXCEPTION)


if __name__ == "__main__":
    # stuff only to run when not called via 'import' here
    try:
        main(sys.argv[1:])
    except SystemExit:
        raise
    except:
        print_exc()
        sys.exit(IsoQuantExitCode.UNCAUGHT_EXCEPTION)



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
import io
from src import serialization
from src.serialization import *


@pytest.fixture
def setup_class(request):
    request.cls.filehandler = io.BytesIO()

    def teardown():
        request.cls.filehandler.close()

    request.addfinalizer(teardown)


@pytest.mark.usefixtures("setup_class")
class TestSerialization:
    @pytest.mark.parametrize(
        "func_wr, func_rd, value",
        [
            (write_int, read_int, 1),
            (write_int, read_int, 0),
            (write_int, read_int, 65536),
            (write_short_int, read_short_int, 0),
            (write_short_int, read_short_int, 65535),
            (write_string, read_string, ""),
            (write_string, read_string, "quite a long string dont you think"),
            (write_string_or_none, read_string_or_none, None),
            (write_string_or_none, read_string_or_none, "None"),
            (write_int_neg, read_int_neg, -1),
            (write_int_neg, read_int_neg, -65536),
            (write_bool_array, read_bool_array, [False]),
            (write_dict, read_dict, {'key1': 'value1', 'key2': 'value2'}),
            (write_dict, read_dict, {'key1': 0, 'key2': (655536, 1)}),
            (write_dict, read_dict, {})
        ]
    )
    def test_write_read_positive(self, value, func_wr, func_rd):
        func_wr(value, self.filehandler)
        self.filehandler.flush()
        self.filehandler.seek(0)
        actual = func_rd(self.filehandler)
        assert actual == value

    @pytest.mark.parametrize(
        "func_wr, func_rd, value",
        [
            (write_int, read_int, [65539, 0, 1]),
            (write_int, read_int, []),
            (write_string_or_none, read_string_or_none, ["str", "", None]),
            (write_string_or_none, read_string_or_none, []),
            (write_string, read_string, ["str", "quite a long string dont you think", ""]),
            (write_string, read_string, []),
            (write_int_neg, read_int_neg, [-6, -33333, 0]),
            (write_int_neg, read_int_neg, [])
        ]
    )
    def test_write_read_lists_positive(self, func_wr, func_rd, value):
        write_list(value, self.filehandler, func_wr)
        self.filehandler.flush()
        self.filehandler.seek(0)
        actual = read_list(self.filehandler, func_rd)
        assert actual == value

    @pytest.mark.parametrize(
        "value",
        [
            ([]),
            ([True, False]),
            ([False, False, False, True, False, True, True, False]),
        ]
    )
    def test_write_read_bool_array_positive(self, value):
        write_bool_array(value, self.filehandler)
        self.filehandler.flush()
        self.filehandler.seek(0)
        actual = read_bool_array(self.filehandler, len(value))
        assert actual == value

    @pytest.mark.parametrize(
        "func_wr, func_rd, list_of_pairs",
        [
            (write_int, read_int, []),
            (write_int, read_int, [(1225, 78854)]),
            (write_int, read_int, [(1225, 78854), (1, 0), (65536, 65536)]),
            (write_string, read_string, [("1225", "78854"), ("", " "), ("65536", "65536")])
        ]
    )
    def test_write_read_list_of_pairs_positive(self, func_wr, func_rd, list_of_pairs):
        write_list_of_pairs(list_of_pairs, self.filehandler, func_wr)
        self.filehandler.flush()
        self.filehandler.seek(0)
        actual = read_list_of_pairs(self.filehandler, func_rd)
        assert actual == list_of_pairs

    def test_read_write_list_with_custom_serializer(self):
        class TestClass:
            def __init__(self, value):
                self.value = value

            def serialize(self, outfile):
                write_int(self.value, outfile)

            @staticmethod
            def deserialize(infile):
                obj = TestClass(0)
                obj.value = read_int(infile)
                return obj

        test_objects = [TestClass(1), TestClass(2), TestClass(3)]
        write_list(test_objects, self.filehandler, TestClass.serialize)
        self.filehandler.flush()
        self.filehandler.seek(0)
        read_objects = read_list(self.filehandler, TestClass.deserialize)

        assert len(test_objects) == len(read_objects)
        for orig, read in zip(test_objects, read_objects):
            assert orig.value == read.value

    def test_termination_int(self):
        write_int(TERMINATION_INT, self.filehandler)
        self.filehandler.flush()
        self.filehandler.seek(0)
        value = read_int(self.filehandler)
        assert value == TERMINATION_INT




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging

from .common import get_read_blocks
from .polya_finder import PolyAInfo
from .polya_verification import shift_polya, shift_polyt

logger = logging.getLogger('IsoQuant')


class AlignmentInfo:
    REGION_TO_CHECK_LEN = 12

    def __init__(self, alignment):
        self.alignment = alignment
        # concat indels
        self.read_exons, self.read_blocks, self.cigar_blocks = get_read_blocks(alignment.reference_start,
                                                                               alignment.cigartuples)
        self.aligned_pairs = None
        self.aligned_pairs_start_index = None
        self.aligned_pairs_end_index = None
        if not self.read_exons:
            return
        self.read_start = self.read_exons[0][0]
        self.read_end = self.read_exons[-1][1]
        self.polya_info = PolyAInfo(-1, -1, -1, -1)
        self.exons_changed = False
        self.cage_hits = []
        self.combined_profile = None

    def construct_profiles(self, profile_constructor):
        self.combined_profile = profile_constructor.construct_profiles(self.read_exons, self.polya_info, self.cage_hits)

    def set_aligned_pairs(self):
        self.aligned_pairs = self.alignment.get_aligned_pairs()
        # indices are created for faster access of aligned_pairs near splice sites
        self.aligned_pairs_start_index = []
        self.aligned_pairs_end_index = []
        exon_index = 0
        for i, (read_pos, ref_pos) in enumerate(self.aligned_pairs):
            if ref_pos == self.read_exons[exon_index][0]:
                self.aligned_pairs_start_index.append(i)
            if ref_pos == self.read_exons[exon_index][1]:
                self.aligned_pairs_end_index.append(i)
                exon_index += 1
            if exon_index == len(self.read_exons):
                break
        # terminal index
        self.aligned_pairs_start_index.append(len(self.aligned_pairs))
        self.aligned_pairs_start_index.append(len(self.aligned_pairs))

    def get_error_count(self, ref_start, ref_end, intron_index=None, left_site=True, chr_record=None):
        if not self.aligned_pairs:
            self.set_aligned_pairs()

        if intron_index is None:
            selected_pairs = self.aligned_pairs
        elif left_site:
            # find aligned pairs near intron start (previous exon end)
            selected_pairs = self.aligned_pairs[
                             max(0, self.aligned_pairs_end_index[intron_index] - AlignmentInfo.REGION_TO_CHECK_LEN):
                             self.aligned_pairs_end_index[intron_index] + AlignmentInfo.REGION_TO_CHECK_LEN]
        else:
            # find aligned pairs near intron end (next exon start)
            selected_pairs = self.aligned_pairs[
                             max(0, self.aligned_pairs_start_index[intron_index + 1] - AlignmentInfo.REGION_TO_CHECK_LEN):
                             self.aligned_pairs_start_index[intron_index + 1] + AlignmentInfo.REGION_TO_CHECK_LEN]

        indel_count = 0
        mismatch_count = 0
        last_ref_pos = 0
        for (read_pos, ref_pos) in selected_pairs:
            # note that ref_pos are 0 based from BAM file, but ref_start and ref_end are 1-based
            true_ref_pos = ref_pos if ref_pos is None else ref_pos + 1
            if true_ref_pos is None:
                # insertion, check previous position is inside or adjacent to the region
                if ref_start - 1 <= last_ref_pos <= ref_end:
                    indel_count += 1
            elif ref_start <= true_ref_pos <= ref_end:
                if read_pos is None:
                    indel_count += 1
                elif chr_record and self.alignment.query_sequence and self.alignment.query_sequence[read_pos] != chr_record[ref_pos]:
                    # chr_record is also 0-based
                    mismatch_count += 1
            last_ref_pos = true_ref_pos or last_ref_pos
            if last_ref_pos > ref_end:
                # no need to check further
                break
        return indel_count, mismatch_count

    def get_seq(self, ref_start, ref_end, intron_region=None):
        assert False
        if not self.aligned_pairs:
            self.set_aligned_pairs()

        if intron_region is None:
            selected_pairs = self.aligned_pairs
        else:
            selected_pairs = self.aligned_pairs[self.aligned_pairs_index[intron_region[0]]:
                                                self.aligned_pairs_index[intron_region[1]+2]]
        seq = ''
        last_ref_pos = 0
        for read_pos, ref_pos in selected_pairs:
            if ref_start <= last_ref_pos <= ref_end:
                if read_pos is not None:
                    seq += self.alignment.seq[read_pos]
            last_ref_pos = ref_pos or last_ref_pos
        return seq

    def add_cage_info(self, cage_finder):
        self.cage_hits = cage_finder.find_cage_peak(self.alignment)

    def add_polya_info(self, polya_finder, polya_fixer):
        self.polya_info = polya_finder.detect_polya(self.alignment)
        polya_exon_count, polyt_exon_count = polya_fixer.correct_read_info(self.read_exons, self.polya_info)

        if polya_exon_count > 0:
            self.polya_info.internal_polya_pos = shift_polya(self.read_exons, polya_exon_count,
                                                             self.polya_info.internal_polya_pos)
            self.polya_info.external_polya_pos = shift_polya(self.read_exons, polya_exon_count,
                                                             self.polya_info.external_polya_pos)
            self.read_exons = self.read_exons[:-polya_exon_count]
            self.read_blocks = self.read_blocks[:-polya_exon_count]
            self.cigar_blocks = self.cigar_blocks[:-polya_exon_count]
            logger.debug("Trimming polyA exons %d: %s" % (polya_exon_count, str(self.read_exons)))
            self.exons_changed = True

        if polyt_exon_count > 0:
            self.polya_info.internal_polyt_pos = shift_polyt(self.read_exons, polyt_exon_count,
                                                             self.polya_info.internal_polyt_pos)
            self.polya_info.external_polyt_pos = shift_polyt(self.read_exons, polyt_exon_count,
                                                             self.polya_info.external_polyt_pos)
            self.read_exons = self.read_exons[polyt_exon_count:]
            self.read_blocks = self.read_blocks[polyt_exon_count:]
            self.cigar_blocks = self.cigar_blocks[polyt_exon_count:]
            logger.debug("Trimming polyT exons %d: %s" % (polyt_exon_count, str(self.read_exons)))
            self.exons_changed = True

        if self.exons_changed:
            self.read_start = self.read_exons[0][0]
            self.read_end = self.read_exons[-1][1]



############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
K-mer indexers for fast approximate string matching in barcode calling.

Classes:
    KmerIndexer: Basic dictionary-based indexer
    ArrayKmerIndexer: Array-based indexer for small k values
    Dict2BitKmerIndexer: Memory-efficient dict with integer keys
    Array2BitKmerIndexer: Memory-efficient array for large barcode sets
    SharedMemoryArray2BitKmerIndexer: Shared memory version for parallel processing
"""

from .base import KmerIndexer, ArrayKmerIndexer
from .two_bit import Dict2BitKmerIndexer, Array2BitKmerIndexer
from .shared_memory import SharedMemoryArray2BitKmerIndexer, SharedMemoryIndexInfo

__all__ = [
    'KmerIndexer',
    'ArrayKmerIndexer',
    'Dict2BitKmerIndexer',
    'Array2BitKmerIndexer',
    'SharedMemoryArray2BitKmerIndexer',
    'SharedMemoryIndexInfo',
]



#!/usr/bin/env python3

from src.post_process import OutputConfig, DictionaryBuilder
from src.plot_output import PlotOutput
import argparse
from src.process_dict import simplify_and_sum_transcripts
from src.gene_model import rank_and_visualize_genes


class FindGenesAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values is None:
            values = 100  # Default value when the flag is used without a value
        setattr(namespace, self.dest, values)


def parse_arguments():
    parser = argparse.ArgumentParser(description="Visualize your IsoQuant output.")
    parser.add_argument(
        "output_directory", type=str, help="Directory containing IsoQuant output files."
    )
    parser.add_argument(
        "--viz_output",
        type=str,
        help="Optional directory to save visualization output files, defaults to the main output directory.",
        default=None,
    )
    parser.add_argument(
        "--gtf",
        type=str,
        help="Optional path to a GTF file if unable to be extracted from IsoQuant log",
        default=None,
    )
    parser.add_argument(
        "--counts", action="store_true", help="Use counts instead of TPM files."
    )
    parser.add_argument(
        "--ref_only",
        action="store_true",
        help="Use only reference transcript quantification instead of transcript model quantification.",
    )
    parser.add_argument(
        "--filter_transcripts",
        type=float,
        help="Filter transcripts by minimum value occurring in at least one condition.",
        default=None,
    )
    parser.add_argument(
        "--gene_list",
        type=str,
        required=True,
        help="Path to a .txt file containing a list of genes, each on its own line.",
    )
    parser.add_argument(
        "--find_genes",
        nargs="?",
        const=100,
        type=int,
        help="Find genes with the highest combined rank and visualize them. Optionally specify the number of top genes to evaluate (default is 100).",
    )
    parser.add_argument(
        "--known_genes_path",
        type=str,
        help="Path to a CSV file containing known target genes.",
        default=None,
    )
    return parser.parse_args()


def main():
    args = parse_arguments()
    output = OutputConfig(
        args.output_directory,
        use_counts=args.counts,
        ref_only=args.ref_only,
        gtf=args.gtf,
    )
    dictionary_builder = DictionaryBuilder(output)
    gene_list = dictionary_builder.read_gene_list(args.gene_list)
    update_names = not all(gene.startswith("ENS") for gene in gene_list)
    gene_dict = dictionary_builder.build_gene_transcript_exon_dictionaries()
    reads_and_class = (
        dictionary_builder.build_read_assignment_and_classification_dictionaries()
    )

    if output.conditions:
        gene_file = (
            output.gene_grouped_tpm
            if not output.use_counts
            else output.gene_grouped_counts
        )
    else:
        gene_file = output.gene_tpm if not output.use_counts else output.gene_counts

    updated_gene_dict = dictionary_builder.update_gene_dict(gene_dict, gene_file)
    if update_names:
        print("Updating Ensembl IDs to gene symbols.")
        updated_gene_dict = dictionary_builder.update_gene_names(updated_gene_dict)

    if output.ref_only or not output.extended_annotation:
        print("Using reference-only based quantification.")
        if output.conditions:
            updated_gene_dict = dictionary_builder.update_transcript_values(
                updated_gene_dict,
                (
                    output.transcript_grouped_tpm
                    if not output.use_counts
                    else output.transcript_grouped_counts
                ),
            )
        else:
            updated_gene_dict = dictionary_builder.update_transcript_values(
                updated_gene_dict,
                (
                    output.transcript_tpm
                    if not output.use_counts
                    else output.transcript_counts
                ),
            )
    else:
        print("Using transcript model quantification.")
        if output.conditions:
            updated_gene_dict = dictionary_builder.update_transcript_values(
                updated_gene_dict,
                (
                    output.transcript_model_grouped_tpm
                    if not output.use_counts
                    else output.transcript_model_grouped_counts
                ),
            )
        else:
            updated_gene_dict = dictionary_builder.update_transcript_values(
                updated_gene_dict,
                (
                    output.transcript_model_tpm
                    if not output.use_counts
                    else output.transcript_model_counts
                ),
            )

    if args.filter_transcripts is not None:
        print(
            f"Filtering transcripts with minimum value {args.filter_transcripts} in at least one condition."
        )
        updated_gene_dict = dictionary_builder.filter_transcripts_by_minimum_value(
            updated_gene_dict, min_value=args.filter_transcripts
        )
    else:
        updated_gene_dict = dictionary_builder.filter_transcripts_by_minimum_value(
            updated_gene_dict
        )

    # Visualization output directory decision
    viz_output_directory = args.viz_output if args.viz_output else args.output_directory

    if args.find_genes:
        print("Finding genes.")
        simple_gene_dict = simplify_and_sum_transcripts(updated_gene_dict)
        path = rank_and_visualize_genes(
            simple_gene_dict,
            viz_output_directory,
            args.find_genes,
            known_genes_path=args.known_genes_path,
        )
        gene_list = dictionary_builder.read_gene_list(path)

    # dictionary_builder.save_gene_dict_to_json(updated_gene_dict, viz_output_directory)
    plot_output = PlotOutput(
        updated_gene_dict,
        gene_list,
        viz_output_directory,
        create_visualization_subdir=(viz_output_directory == args.output_directory),
        reads_and_class=reads_and_class,
        filter_transcripts=args.filter_transcripts,
        conditions=output.conditions,
        use_counts=args.counts,
    )
    plot_output.plot_transcript_map()
    plot_output.plot_transcript_usage()
    plot_output.make_pie_charts()


if __name__ == "__main__":
    main()




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import os
import re
import shutil
from collections import defaultdict

from .common import  rreplace

logger = logging.getLogger('IsoQuant')


def merge_file_list(fname, label, chr_ids):
    return [rreplace(fname, label, f"{label}_{chr_id}") for chr_id in chr_ids]


def merge_files(file_name, label, chr_ids, merged_file_handler, copy_header=True, header_lines=None):
    file_names = merge_file_list(file_name, label, chr_ids)
    file_names.sort(key=lambda s: [int(t) if t.isdigit() else t.lower() for t in re.split(r"(\d+)", s)])
    for i, file_name in enumerate(file_names):
        if not os.path.exists(file_name): continue
        if header_lines is not None:
            header_count = header_lines
        else:
            header_count = 0
            with open(file_name, 'r') as f:
                while f.readline().startswith("#"):
                    header_count += 1
        with open(file_name, 'rt') as f:
            if not (copy_header and i == 0):
                for j in range(header_count):
                    f.readline()
            shutil.copyfileobj(f, merged_file_handler)
    for file_name in file_names:
        os.remove(file_name)


def merge_counts(counter, label, chr_ids, unaligned_reads=0):
    file_name = counter.output_counts_file_name
    merged_file_handler = counter.get_output_file_handler()
    merge_files(file_name, label, chr_ids, merged_file_handler, header_lines=1)

    if counter.usable_file_name:
        for f in merge_file_list(counter.usable_file_name, label, chr_ids):
            counter.load_usable(f)
            os.remove(f)

    stat_dict = defaultdict(int)
    if counter.output_stats_file_name and counter.ignore_read_groups:
        stats_file_names = merge_file_list(counter.output_stats_file_name, label, chr_ids)
        for file_name in stats_file_names:
            for l in open(file_name):
                v = l.strip().split()
                stat_dict[v[0]] += int(v[1])
            os.remove(file_name)

        if unaligned_reads > 0:
            stat_dict["__not_aligned"] = unaligned_reads
        for v in stat_dict.keys():
            merged_file_handler.write("%s\t%d\n" % (v, stat_dict[v]))


def normalize_path(config_path, file_path):
    if os.path.isabs(file_path):
        return os.path.normpath(file_path)
    else:
        return os.path.normpath(os.path.join(os.path.dirname(config_path), file_path))




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################


import gc
import gzip
import logging
import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor

import pysam

from .common import setup_worker_logging, _get_log_params

logger = logging.getLogger('IsoQuant')


def get_chromosome_read_counts(bam_files, chromosomes):
    """
    Quickly get read counts per chromosome from BAM indices.
    Uses pysam.get_index_statistics() which reads from .bai without scanning BAM.

    Returns: dict mapping chr_id -> total_read_count
    """
    chr_counts = {chr_id: 0 for chr_id in chromosomes}

    for bam_file in bam_files:
        try:
            bam = pysam.AlignmentFile(bam_file, "rb")
            try:
                # Get statistics from index
                stats = bam.get_index_statistics()
                for stat in stats:
                    chr_id = stat.contig
                    if chr_id in chr_counts:
                        # Use mapped + unmapped counts
                        chr_counts[chr_id] += stat.mapped + stat.unmapped
            finally:
                bam.close()
        except Exception as e:
            logger.warning(f"Could not read index statistics from {bam_file}: {e}")
            # Fallback: estimate from BAM header
            try:
                bam = pysam.AlignmentFile(bam_file, "rb")
                for chr_id in chromosomes:
                    if chr_id in bam.references:
                        # Rough estimate: count via fetch (slower but works without index stats)
                        try:
                            count = bam.count(chr_id)
                            chr_counts[chr_id] += count
                        except:
                            pass
                bam.close()
            except:
                pass

    return chr_counts


def distribute_chromosomes_weighted(chromosomes, num_workers, chr_read_counts):
    """
    Distribute chromosomes to workers using greedy bin packing.
    Balances load by assigning large chromosomes first.

    Args:
        chromosomes: list of chromosome IDs
        num_workers: number of worker processes
        chr_read_counts: dict mapping chr_id -> read_count

    Returns:
        List of lists, where workers[i] contains chromosome IDs for worker i
    """
    if not chromosomes:
        return [[] for _ in range(num_workers)]

    # Sort chromosomes by read count (descending)
    sorted_chrs = sorted(chromosomes, key=lambda c: chr_read_counts.get(c, 0), reverse=True)

    # Initialize workers
    workers = [[] for _ in range(num_workers)]
    worker_loads = [0] * num_workers

    # Greedy assignment: assign each chromosome to least-loaded worker
    for chr_id in sorted_chrs:
        min_worker_idx = worker_loads.index(min(worker_loads))
        workers[min_worker_idx].append(chr_id)
        worker_loads[min_worker_idx] += chr_read_counts.get(chr_id, 0)

    # Log distribution for debugging
    for i, (chrs, load) in enumerate(zip(workers, worker_loads)):
        if chrs:
            logger.debug(f"Worker {i}: {len(chrs)} chromosomes, {load:,} reads")

    return workers


def collect_chromosome_reads(chr_id, bam_files):
    """
    Collect all read IDs for a specific chromosome by scanning BAM files.
    """
    read_ids = set()
    for bam_file in bam_files:
        try:
            bam = pysam.AlignmentFile(bam_file, "rb")
            try:
                for read in bam.fetch(chr_id):
                    read_ids.add(read.query_name)
            finally:
                bam.close()
        except Exception as e:
            logger.warning(f"Could not fetch reads from {bam_file} for chromosome {chr_id}: {e}")

    return read_ids


def process_table_for_chromosomes(worker_id, input_tsvs, my_chromosomes, bam_files,
                                  output_files, read_column, group_columns, delim):
    """
    Worker function: processes entire table for assigned chromosomes.

    Each worker:
    1. Builds read ID cache for its assigned chromosomes
    2. Streams through table line-by-line (memory efficient)
    3. Writes matching reads to chromosome-specific output files

    Note: Output files preserve all columns from 0 to max(group_columns) to keep
    column indices valid. This avoids index shifting when non-contiguous columns
    are specified (e.g., columns 0 and 2 but not 1).

    Args:
        worker_id: Worker identifier for logging
        input_tsvs: TSV file(s) to read from
        my_chromosomes: List of chromosome IDs assigned to this worker
        bam_files: List of BAM files to scan for read IDs
        output_files: Dict mapping chr_id -> output_file_path
        read_column: Column index for read ID
        group_columns: Tuple of column indices for group values
        delim: Column delimiter
    """
    if not my_chromosomes:
        return 0, 0

    logger.debug(f"Worker {worker_id}: processing {len(my_chromosomes)} chromosomes: {', '.join(my_chromosomes)}")

    # Step 1: Build read ID cache for assigned chromosomes
    logger.debug(f"Worker {worker_id}: building read ID cache...")
    read_cache = {}
    for chr_id in my_chromosomes:
        read_cache[chr_id] = collect_chromosome_reads(chr_id, bam_files)
        logger.debug(f"Worker {worker_id}: cached {len(read_cache[chr_id]):,} reads for {chr_id}")

    # Step 2: Open output files
    out_handles = {}
    for chr_id in my_chromosomes:
        out_handles[chr_id] = open(output_files[chr_id], 'w')

    # Step 3: Stream through table line-by-line
    total_reads_processed = 0
    total_reads_written = 0
    # Keep all columns from 0 to max index to preserve column indices
    max_column_index = max(read_column, max(group_columns))
    min_columns = max_column_index + 1

    try:
        input_files = input_tsvs if isinstance(input_tsvs, list) else [input_tsvs]

        for input_file in input_files:
            # Handle gzipped files
            _, ext = os.path.splitext(input_file)
            if ext.lower() in ['.gz', '.gzip']:
                file_handle = gzip.open(input_file, 'rt')
            else:
                file_handle = open(input_file)

            try:
                for line in file_handle:
                    if line.startswith("#") or line.startswith("read_id\t") or not line.strip():
                        continue

                    total_reads_processed += 1
                    columns = line.rstrip('\n').split(delim)

                    if len(columns) < min_columns:
                        continue

                    read_id = columns[read_column]

                    # Check if this read belongs to any of my chromosomes
                    for chr_id in my_chromosomes:
                        if read_id in read_cache[chr_id]:
                            # Keep all columns from 0 to max_column_index to preserve indices
                            # This ensures column references remain valid in split files
                            output_cols = delim.join(columns[:max_column_index + 1])
                            out_handles[chr_id].write(f"{output_cols}\n")
                            total_reads_written += 1
                            break  # Read can only be on one chromosome
            finally:
                file_handle.close()
    finally:
        # Close output files
        for f in out_handles.values():
            f.close()

    logger.debug(f"Worker {worker_id}: processed {total_reads_processed:,} table entries, "
                 f"wrote {total_reads_written:,} reads")

    return total_reads_processed, total_reads_written


def split_read_table_parallel(sample, input_tsvs, split_reads_file_names, num_threads,
                              read_column=0, group_columns=(1,), delim='\t'):
    """
    Improved parallel table splitting algorithm.

    Strategy: Assign chromosomes to workers, each worker streams table line-by-line for its chromosomes.

    Advantages over old algorithm:
    - No redundant work (each line processed once by each worker)
    - No chunking overhead (line-by-line streaming)
    - Minimal memory usage (no intermediate dicts)
    - Better memory distribution (each worker caches only its chromosomes)
    - True parallelism (no generator bottleneck)

    Args:
        sample: Sample object with file_list
        input_tsvs: TSV file(s) containing read groups
        split_reads_file_names: Dict mapping chr_id -> output_file_path
        num_threads: Number of worker processes
        read_column: Column index for read ID (default: 0)
        group_columns: Tuple of column indices for group values (default: (1,))
        delim: Column delimiter (default: '\t')
    """
    logger.info(f"Splitting table {input_tsvs} across {num_threads} workers")

    bam_files = list(map(lambda x: x[0], sample.file_list))
    chromosomes = list(split_reads_file_names.keys())

    if not chromosomes:
        logger.warning("No chromosomes to process")
        return

    # Step 1: Get chromosome read counts from BAM indices for load balancing
    logger.info("Analyzing chromosome sizes from BAM indices...")
    chr_read_counts = get_chromosome_read_counts(bam_files, chromosomes)

    total_reads = sum(chr_read_counts.values())
    logger.info(f"Total reads across {len(chromosomes)} chromosomes: {total_reads:,}")

    # Step 2: Distribute chromosomes to workers (weighted by read count)
    num_workers = min(num_threads, len(chromosomes))  # No point having more workers than chromosomes
    chr_assignments = distribute_chromosomes_weighted(chromosomes, num_workers, chr_read_counts)

    # Filter out empty assignments
    chr_assignments = [chrs for chrs in chr_assignments if chrs]
    num_workers = len(chr_assignments)

    logger.info(f"Using {num_workers} workers to process {len(chromosomes)} chromosomes")

    # Step 3: Initialize output files (truncate)
    for chr_id in chromosomes:
        open(split_reads_file_names[chr_id], 'w').close()

    # Step 4: Launch workers
    # Clean up parent memory before spawning workers
    gc.collect()
    mp_context = multiprocessing.get_context('spawn')
    log_file, log_level = _get_log_params()
    with ProcessPoolExecutor(max_workers=num_workers, mp_context=mp_context,
                             initializer=setup_worker_logging,
                             initargs=(log_file, log_level)) as executor:
        futures = []
        for worker_id, my_chrs in enumerate(chr_assignments):
            future = executor.submit(
                process_table_for_chromosomes,
                worker_id,
                input_tsvs,
                my_chrs,
                bam_files,
                split_reads_file_names,
                read_column,
                group_columns,
                delim
            )
            futures.append(future)

        # Wait for all workers and collect results
        total_processed = 0
        total_written = 0
        for future in futures:
            processed, written = future.result()
            total_processed += processed
            total_written += written

    logger.info(f"Table splitting complete: {total_written:,} reads written to {len(chromosomes)} chromosome files")



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import os

import gffutils
import pysam
from pyfaidx import Fasta

from .stats import EnumStats
from .alignment_processor import AlignmentCollector
from .assignment_io import IOSupport, TmpFileAssignmentPrinter, SqantiTSVPrinter
from .assignment_loader import create_assignment_loader, BasicReadAssignmentLoader
from .barcode_calling.umi_filtering import create_transcript_info_dict, UMIFilter
from .file_naming import (
    read_groups_file_name,
    bamstat_file_name,
    read_group_file_from_saves,
    dynamic_pools_file_name,
    saves_file_name,
    reads_processed_lock_file_name,
    read_stat_file_name,
    transcript_stat_file_name,
    umi_filtered_lock_file_name,
    allinfo_file_name,
    allinfo_stats_file_name,
)
from .gene_info import TranscriptModelType
from .graph_based_model_construction import GraphBasedModelConstructor
from .id_policy import SimpleIDDistributor, ExcludingIdDistributor, FeatureIdStorage
from .read_groups import create_read_grouper, get_grouping_strategy_names, load_table
from .transcript_printer import GFFPrinter, VoidTranscriptPrinter, create_extended_storage
from .assignment_aggregator import ReadAssignmentAggregator
from .string_pools import setup_string_pools
from .common import large_output_enabled

logger = logging.getLogger('IsoQuant')


# Helper functions to reduce code duplication

def load_dynamic_pools(string_pools, dynamic_pools_file):
    """Load dynamic pools from file if it exists."""
    if os.path.exists(dynamic_pools_file):
        logger.debug(f"Loading dynamic pools from {dynamic_pools_file}")
        with open(dynamic_pools_file, 'rb') as f:
            string_pools.deserialize_dynamic_pools(f)


def save_dynamic_pools(string_pools, dynamic_pools_file):
    """Save dynamic pools to file if they have data."""
    if string_pools.has_dynamic_pools():
        logger.debug(f"Saving dynamic pools to {dynamic_pools_file}")
        with open(dynamic_pools_file, 'wb') as f:
            string_pools.serialize_dynamic_pools(f)


def save_read_groups(group_file, read_groups):
    """Save read_groups to file (handles both single set and list of sets)."""
    with open(group_file, "w") as group_dump:
        if isinstance(read_groups, list):
            # MultiReadGrouper: list of sets
            group_dump.write("%d\n" % len(read_groups))
            for group_set in read_groups:
                group_dump.write("%s\n" % ";".join(sorted(group_set)))
        else:
            # Single grouper: single set
            group_dump.write("1\n")
            group_dump.write("%s\n" % ";".join(sorted(read_groups)))


def load_read_groups(group_file, read_grouper):
    """Load read_groups from file into read_grouper."""
    with open(group_file) as f:
        lines = [line.strip() for line in f]
        num_strategies = int(lines[0])
        if isinstance(read_grouper.read_groups, list):
            # MultiReadGrouper: list of sets
            read_grouper.read_groups = []
            for i in range(num_strategies):
                groups = set(lines[i + 1].split(";")) if lines[i + 1] else set()
                read_grouper.read_groups.append(groups)
        else:
            # Single grouper: single set
            groups_str = lines[1] if len(lines) > 1 else ""
            read_grouper.read_groups = set(groups_str.split(";")) if groups_str else set()


def load_barcode_dict(sample, chr_id):
    """Load barcode dictionary for a chromosome."""
    barcode_dict = {}
    if sample.barcodes_split_reads:
        barcode_file = sample.get_barcodes_split_file(chr_id)
        if os.path.exists(barcode_file):
            logger.debug(f"Loading barcodes from {barcode_file}")
            for line in open(barcode_file):
                if line.startswith("#") or line.startswith("read_id\t"):
                    continue
                parts = line.strip().split('\t')
                if len(parts) >= 3:
                    barcode = parts[1]
                    if barcode == '*' or not barcode:
                        continue
                    umi = parts[2] if parts[2] else None
                    barcode_dict[parts[0]] = (barcode, umi)
            logger.debug("Loaded %d barcodes" % len(barcode_dict))
    return barcode_dict


def collect_reads_in_parallel(sample, chr_id, chr_ids, args, processed_read_manager_type):
    current_chr_record = Fasta(args.reference, indexname=args.fai_file_name)[chr_id]
    if args.high_memory:
        current_chr_record = str(current_chr_record)
    read_grouper = create_read_grouper(args, sample, chr_id)
    lock_file = sample.get_collected_lock_file(chr_id)
    save_file = sample.get_save_file(chr_id)
    group_file = read_groups_file_name(save_file)
    bamstat_file = bamstat_file_name(save_file)
    processed_reads_manager = processed_read_manager_type(sample, args.multimap_strategy, chr_ids, args.genedb)

    if os.path.exists(lock_file) and args.resume:
        logger.info("Detected processed reads for " + chr_id)
        if os.path.exists(group_file) and os.path.exists(save_file):
            load_read_groups(group_file, read_grouper)
            alignment_stat_counter = EnumStats(bamstat_file)

            # Build string pools for loading serialized data
            gffutils_db = gffutils.FeatureDB(args.genedb) if args.genedb else None
            string_pools = setup_string_pools(args, sample, chr_ids, chr_id, gffutils_db,
                                              load_barcode_pool=True, load_tsv_pools=True)

            # Load dynamic pools (for read groups from BAM tags/read IDs)
            load_dynamic_pools(string_pools, sample.get_dynamic_pools_file(chr_id))

            loader = BasicReadAssignmentLoader(save_file, string_pools)
            while loader.has_next():
                for read_assignment in loader.get_next():
                    if read_assignment is None: continue
                    processed_reads_manager.load_read(read_assignment)
            logger.info("Loaded data for " + chr_id)
            return chr_id, read_grouper.read_groups, alignment_stat_counter, processed_reads_manager
        else:
            logger.warning("Something is wrong with save files for %s, will process from scratch " % chr_id)
            if not os.path.exists(group_file):
                logger.warning("%s does not exist" % group_file)
            if not os.path.exists(save_file):
                logger.warning("%s does not exist" % save_file)
            os.remove(lock_file)

    if os.path.exists(lock_file):
        os.remove(lock_file)

    tmp_printer = TmpFileAssignmentPrinter(save_file, args)
    bam_files = list(map(lambda x: x[0], sample.file_list))
    bam_file_pairs = [(pysam.AlignmentFile(bam, "rb", require_index=True), bam) for bam in bam_files]
    gffutils_db = gffutils.FeatureDB(args.genedb) if args.genedb else None
    illumina_bam = sample.illumina_bam

    # Build string pools for memory optimization
    string_pools = setup_string_pools(args, sample, chr_ids, chr_id, gffutils_db,
                                      load_barcode_pool=True, load_tsv_pools=True)

    # Load barcode dict for this chromosome if available
    barcode_dict = load_barcode_dict(sample, chr_id)

    logger.info("Processing chromosome " + chr_id)
    alignment_collector = \
        AlignmentCollector(chr_id, bam_file_pairs, args, illumina_bam, gffutils_db, current_chr_record, read_grouper,
                           barcode_dict, args.max_coverage_small_chr, args.max_coverage_normal_chr, string_pools)

    for gene_info, assignment_storage in alignment_collector.process():
        tmp_printer.add_gene_info(gene_info)
        for read_assignment in assignment_storage:
            tmp_printer.add_read_info(read_assignment)
            processed_reads_manager.add_read(read_assignment)
    save_read_groups(group_file, read_grouper.read_groups)
    alignment_collector.alignment_stat_counter.dump(bamstat_file)

    for bam in bam_file_pairs:
        bam[0].close()

    tmp_printer.close()

    # Save dynamic pools if they have data (for read groups from BAM tags/read IDs)
    save_dynamic_pools(string_pools, sample.get_dynamic_pools_file(chr_id))

    # Save barcode/UMI pools for --barcoded_bam mode (pools built on-the-fly from BAM tags)
    if getattr(args, 'barcoded_bam', False) and len(string_pools.barcode_pool) > 0:
        string_pools.save_barcode_umi_pools(sample.get_barcode_pools_file(chr_id))

    processed_reads_manager.finalize(chr_id)
    logger.info("Finished processing chromosome " + chr_id)
    open(lock_file, "w").close()

    return chr_id, read_grouper.read_groups, alignment_collector.alignment_stat_counter, processed_reads_manager


def construct_models_in_parallel(sample, chr_id, chr_ids, saves_prefix, args, read_groups):
    from .read_groups import get_grouping_pool_types
    logger.info("Processing chromosome " + chr_id)
    use_filtered_reads = args.mode.needs_pcr_deduplication()

    # Derive read_group_file prefix from saves_prefix
    read_group_file_prefix = read_group_file_from_saves(saves_prefix)

    # Check if barcode pool is needed for any grouper
    pool_types = get_grouping_pool_types(args)
    needs_barcode_pool = any(pt == 'barcode' for pt in pool_types.values())

    # Build string pools for memory optimization
    string_pools = setup_string_pools(args, sample, chr_ids, chr_id,
                                      load_barcode_pool=needs_barcode_pool, load_tsv_pools=True,
                                      read_group_file_prefix=read_group_file_prefix)

    # Load dynamic pools (for read groups from BAM tags/read IDs)
    load_dynamic_pools(string_pools, dynamic_pools_file_name(saves_prefix, chr_id))

    loader = create_assignment_loader(chr_id, saves_prefix, args.genedb, args.reference, args.fai_file_name, string_pools, use_filtered_reads)

    chr_dump_file = saves_file_name(saves_prefix, chr_id)
    lock_file = reads_processed_lock_file_name(saves_prefix, chr_id)
    chr_read_stat_file = read_stat_file_name(chr_dump_file)
    chr_transcript_stat_file = transcript_stat_file_name(chr_dump_file)
    construct_models = not args.no_model_construction

    if os.path.exists(lock_file):
        if args.resume:
            logger.info("Processed assignments from chromosome " + chr_id + " detected")
            read_stat = EnumStats(chr_read_stat_file)
            transcript_stat = EnumStats(chr_transcript_stat_file) if construct_models else EnumStats()
            return read_stat, transcript_stat
        os.remove(lock_file)

    grouping_strategy_names = get_grouping_strategy_names(args)
    aggregator = ReadAssignmentAggregator(args, sample, string_pools, loader.genedb, chr_id,
                                         grouping_strategy_names=grouping_strategy_names)

    transcript_stat_counter = EnumStats()
    io_support = IOSupport(args)
    transcript_id_distributor = ExcludingIdDistributor(loader.genedb, chr_id)
    exon_id_storage = FeatureIdStorage(SimpleIDDistributor(), loader.genedb, chr_id, "exon")

    # Use chromosome-specific prefix for output files
    chr_prefix = sample.get_chr_prefix(chr_id)

    if construct_models:
        tmp_gff_printer = GFFPrinter(sample.out_dir, chr_prefix, exon_id_storage,
                                     output_r2t=large_output_enabled(args, "read2transcripts"),
                                     check_canonical=args.check_canonical)
    else:
        tmp_gff_printer = VoidTranscriptPrinter()
    if construct_models and args.genedb:
        tmp_extended_gff_printer = GFFPrinter(sample.out_dir, chr_prefix, exon_id_storage,
                                              gtf_suffix=".extended_annotation.gtf",
                                              output_r2t=False, check_canonical=args.check_canonical)
    else:
        tmp_extended_gff_printer = VoidTranscriptPrinter()

    sqanti_t2t_printer = SqantiTSVPrinter(sample.get_t2t_tsv_file(chr_id), args, IOSupport(args)) \
        if args.sqanti_output else VoidTranscriptPrinter()
    novel_model_storage = []

    while loader.has_next():
        gene_info, assignment_storage = loader.get_next()
        logger.debug("Processing %d reads" % len(assignment_storage))
        for read_assignment in assignment_storage:
            if read_assignment is None:
                continue
            aggregator.read_stat_counter.add(read_assignment.assignment_type)
            aggregator.global_printer.add_read_info(read_assignment)
            aggregator.global_counter.add_read_info(read_assignment)

        if construct_models:
            strategy_names = aggregator.grouping_strategy_names if hasattr(aggregator, 'grouping_strategy_names') else []
            model_constructor = GraphBasedModelConstructor(gene_info, loader.chr_record, args,
                                                           aggregator.transcript_model_global_counter,
                                                           aggregator.gene_model_global_counter,
                                                           transcript_id_distributor,
                                                           grouping_strategy_names=strategy_names,
                                                           use_technical_replicas=sample.use_technical_replicas,
                                                           string_pools=string_pools)
            model_constructor.process(assignment_storage)
            if args.check_canonical:
                io_support.add_canonical_info(model_constructor.transcript_model_storage, gene_info)
            tmp_gff_printer.dump(model_constructor.gene_info, model_constructor.transcript_model_storage)
            tmp_gff_printer.dump_read_assignments(model_constructor)
            for m in model_constructor.transcript_model_storage:
                if m.transcript_type != TranscriptModelType.known:
                    novel_model_storage.append(m)
            for a in model_constructor.transcript2transcript:
                sqanti_t2t_printer.add_read_info(a)
            for t in model_constructor.transcript_model_storage:
                transcript_stat_counter.add(t.transcript_type)

    aggregator.global_counter.dump()
    aggregator.read_stat_counter.dump(chr_read_stat_file)
    if construct_models:
        if loader.genedb:
            all_models, gene_info = create_extended_storage(loader.genedb, chr_id, loader.chr_record, novel_model_storage)
            if args.check_canonical:
                io_support.add_canonical_info(all_models, gene_info)
            tmp_extended_gff_printer.dump(gene_info, all_models)
        aggregator.transcript_model_global_counter.dump()
        aggregator.gene_model_global_counter.dump()
        transcript_stat_counter.dump(chr_transcript_stat_file)
    logger.info("Finished processing chromosome " + chr_id)
    open(lock_file, "w").close()

    return aggregator.read_stat_counter, transcript_stat_counter


def filter_umis_in_parallel(sample, chr_id, chr_ids, args, edit_distance, output_filtered_reads=False,
                            barcode_remap=None, output_prefix_override=None):
    transcript_type_dict = create_transcript_info_dict(args.genedb, [chr_id])
    out_prefix = output_prefix_override if output_prefix_override else sample.out_umi_filtered_done
    umi_filtered_done = umi_filtered_lock_file_name(out_prefix, chr_id, edit_distance)
    all_info_file_name = allinfo_file_name(out_prefix, chr_id, edit_distance)
    stats_output_file_name = allinfo_stats_file_name(out_prefix, chr_id, edit_distance)

    if os.path.exists(umi_filtered_done):
        if args.resume:
            return all_info_file_name, stats_output_file_name, umi_filtered_done
        os.remove(umi_filtered_done)

    logger.info("Filtering PCR duplicates for chromosome " + chr_id)
    barcode_feature_table = {}
    if args.barcode2spot:
        from .read_groups import parse_barcode2spot_spec
        filename, barcode_col, spot_cols = parse_barcode2spot_spec(args.barcode2spot)
        with open(filename, 'r') as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                parts = line.split('\t')
                if len(parts) > barcode_col:
                    barcode = parts[barcode_col]
                    # Extract only the specified spot columns
                    cell_types = [parts[col] for col in spot_cols if col < len(parts)]
                    barcode_feature_table[barcode] = cell_types

    # Build string pools for memory optimization
    # Barcode pool is critical: ReadAssignments were serialized with barcode_id values
    # that reference the barcode pool from read collection
    string_pools = setup_string_pools(args, sample, chr_ids, chr_id,
                                      load_barcode_pool=True, load_tsv_pools=False)

    # Load dynamic pools (for read groups from BAM tags/read IDs)
    load_dynamic_pools(string_pools, sample.get_dynamic_pools_file(chr_id))

    # When using barcode_remap (spot-based dedup), never produce filtered reads
    if barcode_remap:
        output_filtered_reads = False
    umi_filter = UMIFilter(args.umi_length, edit_distance, barcode_remap=barcode_remap)
    filtered_reads = sample.get_filtered_reads_file(chr_id) if output_filtered_reads else None
    umi_filter.process_single_chr(chr_id, sample.out_raw_file,
                                  transcript_type_dict,
                                  barcode_feature_table,
                                  all_info_file_name,
                                  filtered_reads,
                                  stats_output_file_name,
                                  string_pools)
    open(umi_filtered_done, "w").close()
    logger.info("PCR duplicates filtered for chromosome " + chr_id)

    return all_info_file_name, stats_output_file_name, umi_filtered_done



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import math
from enum import Enum

from .isoform_assignment import ReadAssignmentType
from .common import intersection_len


logger = logging.getLogger('IsoQuant')


class MultimapResolvingStrategy(Enum):
    ignore_multimapper = 0
    merge = 1
    take_best = 2


class MultimapResolver:
    duplicate_counter = 0

    def __init__(self, strategy):
        self.strategy = strategy

    # list of ShortReadAssignments
    def resolve(self, assignment_list):
        if assignment_list is None or len(assignment_list) <= 1:
            return assignment_list

        if self.strategy == MultimapResolvingStrategy.ignore_multimapper:
            for a in assignment_list:
                a.assignment_type = ReadAssignmentType.suspended
            return assignment_list

        elif self.strategy == MultimapResolvingStrategy.merge:
            return self.merge_assignments(assignment_list)

        elif self.strategy == MultimapResolvingStrategy.take_best:
            return self.select_best_assignment(assignment_list)

        else:
            raise ValueError("Unsupported multimap strategy")

    def select_best_assignment(self, assignment_list):
        logger.debug("Resolving read %s" % assignment_list[0].read_id)
        primary_unique = []
        consistent_assignments = []
        inconsistent_assignments = []
        primary_inconsistent = []
        noninformative = []

        for i, a in enumerate(assignment_list):
            if a.assignment_type.is_inconsistent():
                if not a.multimapper:
                    primary_inconsistent.append(i)
                inconsistent_assignments.append(i)
            elif a.assignment_type.is_consistent():
                consistent_assignments.append(i)
                if not a.multimapper and not a.assignment_type == ReadAssignmentType.ambiguous:
                    primary_unique.append(i)
            else:
                noninformative.append(i)

        if primary_unique:
            return self.filter_assignments(assignment_list, primary_unique)

        if consistent_assignments:
            return self.filter_assignments(assignment_list, consistent_assignments)

        if primary_inconsistent:
            return self.select_best_inconsistent(assignment_list, primary_inconsistent)

        if inconsistent_assignments:
            return self.select_best_inconsistent(assignment_list, inconsistent_assignments)

        if noninformative:
            return self.select_noninformative(assignment_list, noninformative)

        logger.warning("Unexpected assignments in multimap resolution %s" % assignment_list[0].read_id)
        return assignment_list[0:1]

    @staticmethod
    def select_best_inconsistent(assignment_list, inconsistent_assignments):
        if len(inconsistent_assignments) <= 1:
            return MultimapResolver.filter_assignments(assignment_list, inconsistent_assignments)

        assignment_scores = []
        for i in inconsistent_assignments:
            assignment_scores.append((assignment_list[i].penalty_score, i))
        best_score = min(assignment_scores, key=lambda x: x[0])[0]
        best_assignments = [x[1] for x in filter(lambda x: x[0] == best_score, assignment_scores)]
        return MultimapResolver.filter_assignments(assignment_list, best_assignments)

    @staticmethod
    def merge_assignments(assignment_list):
        informative_assignments = set()
        for i in range(len(assignment_list)):
            if assignment_list[i].assignment_type != ReadAssignmentType.noninformative:
                informative_assignments.add(i)

        return MultimapResolver.filter_assignments(assignment_list, informative_assignments)

    # select all given assignments from the list, makr rest as suspended
    # if selected assignments feature more than 1 gene/isoform - mark as ambiguous accordingly
    @staticmethod
    def filter_assignments(assignment_list, assignments_to_keep):
        assignments_to_keep = MultimapResolver.find_duplicates(assignment_list, assignments_to_keep)

        all_genes = set()
        all_isoforms = set()
        for i in assignments_to_keep:
            assignment = assignment_list[i]
            all_genes.update(assignment.genes)
            all_isoforms.update(assignment.isoforms)

        change_transcript_assignment_type = len(all_isoforms) > 1
        change_gene_assignment_type = len(all_genes) > 1

        assignments_to_keep_set = set(assignments_to_keep)
        for i in range(len(assignment_list)):
            assignment = assignment_list[i]
            if i in assignments_to_keep_set:
                if assignment.assignment_type.is_inconsistent():
                    ambiguity_assignment_type = ReadAssignmentType.inconsistent_ambiguous
                else:
                    ambiguity_assignment_type = ReadAssignmentType.ambiguous

                if change_transcript_assignment_type:
                    assignment.assignment_type = ambiguity_assignment_type
                    assignment.multimapper = True
                if change_gene_assignment_type:
                    assignment.gene_assignment_type = ambiguity_assignment_type
                    assignment.multimapper = True
            else:
                assignment.assignment_type = ReadAssignmentType.suspended
                assignment.gene_assignment_type = ReadAssignmentType.suspended

        return assignment_list

    # finds groups of duplicated assignments
    @staticmethod
    def find_duplicates(assignment_list, assignment_indices):
        if len(assignment_indices) <= 1:
            return assignment_indices

        selected_assignments = []
        discarded_duplicates = set()
        for i in range(len(assignment_indices)):
            index1 = assignment_indices[i]
            if index1 in discarded_duplicates:
                # this assignment was already discarded - no need to compare it
                continue
            # if it was not discarded earlier - keep anyway; next assignments will either duplicate it or won't
            selected_assignments.append(index1)

            for j in range(i + 1, len(assignment_indices)):
                # find duplicating assignments among the next onex
                index2 = assignment_indices[j]
                if index2 in discarded_duplicates:
                    # this assignment was already discarded - no need to compare it
                    continue

                if assignment_list[index1] == assignment_list[index2]:
                    # discard duplicating assignment with larger index
                    discarded_duplicates.add(index2)

                    # check if duplicates come from the same region,
                    # if they are - it means we have identical records in the BAM files, print a warning
                    if assignment_list[index1].genomic_region == assignment_list[index2].genomic_region:
                        MultimapResolver.duplicate_counter += 1
                        example_assignment = assignment_list[index1]
                        if MultimapResolver.duplicate_counter == 5:
                            logger.warning(
                                "More possible duplicates were detected but will not be printed. "
                                "Use --debug to see all of them in the isoquant.log."
                                "All duplicated entries will be ignored. "
                                "This maybe caused either by the same read having more than one alignment with "
                                "the same reference id, start and end coordinates, "
                                "or by duplicated read names in the original files. ")
                        if MultimapResolver.duplicate_counter >= 5:
                            logger.debug(
                                "Read %s seems to have duplicated alignment entries at %s: %d and will be ignored. "
                                % (example_assignment.read_id, example_assignment.chr_id, example_assignment.start))
                        else:
                            logger.info(
                                "Read %s seems to have duplicated alignment entries at %s: %d and will be ignored. "
                                % (example_assignment.read_id, example_assignment.chr_id, example_assignment.start))

        return selected_assignments

    @staticmethod
    def select_noninformative(assignment_list, assignment_indices):
        # triplets (overlap_length, genomic_region_start, index)
        overlap_index_list = []
        max_overlap_len = 0

        for i in assignment_indices:
            assignment = assignment_list[i]
            overlap_len = intersection_len(assignment.genomic_region, (assignment.start, assignment.end))
            max_overlap_len = max(overlap_len, max_overlap_len)
            overlap_index_list.append((overlap_len, assignment.genomic_region[0], i))

        # select assignment with the best overlap with genic region and lowest region start (for reproducibility)
        min_region_start = math.inf
        best_assignment = -1
        for info in overlap_index_list:
            if info[0] == max_overlap_len and info[1] < min_region_start:
                min_region_start = info[1]
                best_assignment = info[2]

        assert best_assignment != -1
        return MultimapResolver.filter_assignments(assignment_list, {best_assignment})




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import math

from .common import interval_len, intervals_total_length
from .isoform_assignment import MatchEvent, MatchEventSubtype

logger = logging.getLogger('IsoQuant')


# recalculate polya site considering terminal exons are fake
def shift_polya(read_exons, exon_count, polya_pos):
    if exon_count == 0 or exon_count == len(read_exons) or polya_pos == -1:
        return polya_pos

    dist_to_polya = 0
    for i in range(exon_count):
        exon = read_exons[-i - 1]
        if exon[0] > polya_pos:
            continue
        elif dist_to_polya == 0:
            # no exons counted yet
            dist_to_polya += polya_pos - exon[0]
        else:
            dist_to_polya += interval_len(exon)
    return read_exons[-exon_count - 1][1] + dist_to_polya


# recalculate polyt site considering terminal exons are fake
def shift_polyt(read_exons, exon_count, polyt_pos):
    if exon_count == 0 or exon_count == len(read_exons) or polyt_pos == -1:
        return polyt_pos

    dist_to_polya = 0
    for i in range(exon_count):
        exon = read_exons[i]
        if exon[1] < polyt_pos:
            continue
        elif dist_to_polya == 0:
            # no exons counted yet
            dist_to_polya += exon[1] - polyt_pos
        else:
            dist_to_polya += interval_len(exon)
    return read_exons[exon_count][0] - dist_to_polya


class PolyAFixer:
    def __init__(self, params):
        self.params = params

    def correct_read_info(self, read_exons, polya_info):
        if len(read_exons) == 1:
            return 0, 0

        polya_exon_count = self.count_polya_exons(read_exons, polya_info.internal_polya_pos)
        polyt_exon_count = self.count_polyt_exons(read_exons, polya_info.internal_polyt_pos)

        if polyt_exon_count > 0 and polya_exon_count > 0:
            logger.debug("Both PolyA and PolyT fake terminal exons found: %d, %d" % (polya_exon_count, polyt_exon_count))
        if polyt_exon_count + polya_exon_count == len(read_exons):
            logger.debug("All exons seem to be consist of polyA/T")
            polyt_exon_count -= 1
            polya_exon_count -= 1

        return polya_exon_count, polyt_exon_count

    # correct polyA position when fake terminal exons are present
    def count_polya_exons(self, read_exons, internal_polya_pos):
        # find exons beyond internal polyA
        if internal_polya_pos == -1:
            return 0
        polya_exon_count = 0
        for i in range(len(read_exons)):
            exon = read_exons[-i - 1]
            if exon[1] <= internal_polya_pos:
                break
            len_to_polya = internal_polya_pos - exon[0]
            if len_to_polya <= 0 or \
                    (len_to_polya <= self.params.max_fake_terminal_exon_len and
                     exon[1] - internal_polya_pos > 2 * len_to_polya):
                # more than 2/3 of the exon is polyA
                polya_exon_count += 1

        logger.debug("Detected polya exons %d" % polya_exon_count)
        # correcting fake terminal exons
        return polya_exon_count


    # correct polyT position when fake terminal exons are present or isoform has short terminal exons
    def count_polyt_exons(self, read_exons, internal_polyt_pos):
        # find exons beyond internal polyT
        if internal_polyt_pos == -1:
            return 0
        polya_exon_count = 0
        for i in range(len(read_exons)):
            exon = read_exons[i]
            if exon[0] >= internal_polyt_pos:
                break
            len_to_polyt = exon[1] - internal_polyt_pos
            if len_to_polyt <= 0 or \
                    (len_to_polyt <= self.params.max_fake_terminal_exon_len and
                    2 * len_to_polyt < internal_polyt_pos - exon[0]):
                # more than 2/3 of the exon is polyT
                polya_exon_count += 1

        logger.debug("Detected polyt exons %d" % polya_exon_count)
        # correcting fake terminal exons
        return polya_exon_count


class PolyAVerifier:
    def __init__(self, gene_info, params):
        self.params = params
        self.gene_info = gene_info
        self.polya_fixer = PolyAFixer(params)

    # check consistency with polyA
    def verify_read_ends(self, combined_read_profile, isoform_id, matching_events):
        logger.debug("+ Validating polyA/T sites")
        if isoform_id is None:
            logger.debug("+ No sites found, ciao")
            return matching_events

        logger.debug("+ Checking read ends in isoform %s" % isoform_id)
        read_exons = combined_read_profile.read_split_exon_profile.read_features
        polya_info = combined_read_profile.polya_info
        isoform_exons = self.gene_info.all_isoforms_exons[isoform_id]

        if self.gene_info.isoform_strands[isoform_id] == '+':
            # check for internal intron polyA priming
            matching_events, is_internal = self.check_internal_polya(polya_info.internal_polya_pos, matching_events)
            if not is_internal and (polya_info.external_polya_pos != -1 or polya_info.internal_polya_pos != -1):
                matching_events = self.verify_polya(isoform_exons, read_exons, polya_info, matching_events)

        elif self.gene_info.isoform_strands[isoform_id] == '-':
            # check for internal intron polyT priming
            matching_events, is_internal = self.check_internal_polyt(polya_info.internal_polyt_pos, matching_events)
            if not is_internal and (polya_info.external_polyt_pos != -1 or polya_info.internal_polyt_pos != -1):
                matching_events = self.verify_polyt(isoform_exons, read_exons, polya_info, matching_events)

        if not matching_events:
            logger.debug("Empty event list after polyA verification")
            matching_events = [MatchEvent(MatchEventSubtype.none)]
        return matching_events

    def verify_polya(self, isoform_exons, read_exons, polya_info, matching_events):
        assert polya_info.external_polya_pos != -1 or polya_info.internal_polya_pos != -1
        isoform_end = isoform_exons[-1][1]
        event_to_remove = -1
        fake_terminal_exon_count = 0
        terminal_exon_misaligned = 0
        # checking events
        for i, event in enumerate(matching_events):
            if event.event_type in [MatchEventSubtype.major_exon_elongation_right,
                                    MatchEventSubtype.exon_elongation_right]:
                # substitute major elongation with APA site
                event_to_remove = i
            elif event.event_type == MatchEventSubtype.fake_terminal_exon_right:
                fake_terminal_exon_count += 1
            elif event.event_type == MatchEventSubtype.terminal_exon_misalignment_right:
                terminal_exon_misaligned += 1

        if terminal_exon_misaligned > 0 and fake_terminal_exon_count > 0:
            logger.debug("Odd case, both terminal exon misalignment and fake exons are present")
            logger.debug(isoform_exons)
            logger.debug(read_exons)
        if event_to_remove != -1:
            del matching_events[event_to_remove]

        new_events = self.check_if_close(isoform_end, polya_info.external_polya_pos,  polya_info.internal_polya_pos,
                                         matching_events, MatchEventSubtype.correct_polya_site_right)
        if new_events is not None:
            logger.debug("Original polyA looks good")
            return new_events

        logger.debug("Original polyA seems distant")
        # if terminal exon is fake, move polyA accordingly
        external_polya_pos, internal_polya_pos = \
            self.correct_polya_positions(read_exons, fake_terminal_exon_count, polya_info)
        logger.debug("+ Corrected external %d, corrected internal %d" % (external_polya_pos, internal_polya_pos))

        if terminal_exon_misaligned:
            # we believe isofrom end is a true polyA site
            external_polya_pos = isoform_exons[-1][1]
            internal_polya_pos = isoform_exons[-1][1]
        else:
            matching_events, external_polya_pos, internal_polya_pos = \
                self.detect_reference_exons_beyond_polya(isoform_exons, external_polya_pos, internal_polya_pos, matching_events)

        new_events = self.check_if_close(isoform_end, external_polya_pos, internal_polya_pos,
                                         matching_events, MatchEventSubtype.correct_polya_site_right)
        if new_events is not None:
            logger.debug("Corrected polyA seems good")
            return new_events

        polya_pos = external_polya_pos if internal_polya_pos == -1 else internal_polya_pos
        dist_to_polya = abs(polya_pos - isoform_end)
        logger.debug("+ Distance to polyA is %d" % dist_to_polya)
        if dist_to_polya > self.params.apa_delta:
            logger.debug("+ Seems like APA site")
            matching_events.append(MatchEvent(MatchEventSubtype.alternative_polya_site_right, event_info=polya_pos))
        else:
            logger.debug("+ Seems like correct polyA, odd case")
            matching_events.append(MatchEvent(MatchEventSubtype.correct_polya_site_right, event_info=polya_pos))
        return matching_events

    def verify_polyt(self, isoform_exons, read_exons, polya_info, matching_events):
        assert polya_info.external_polyt_pos != -1 or polya_info.internal_polyt_pos != -1
        isoform_start = isoform_exons[0][0]
        event_to_remove = -1
        fake_terminal_exon_count = 0
        terminal_exon_misaligned = 0

        for i, event in enumerate(matching_events):
            if event.event_type in [MatchEventSubtype.major_exon_elongation_left,
                                    MatchEventSubtype.exon_elongation_left]:
                # substitute elongation with APA site if major
                event_to_remove = i
            elif event.event_type == MatchEventSubtype.fake_terminal_exon_left:
                fake_terminal_exon_count += 1
            elif event.event_type == MatchEventSubtype.terminal_exon_misalignment_left:
                terminal_exon_misaligned += 1

        if terminal_exon_misaligned > 0 and fake_terminal_exon_count > 0:
            logger.debug("Odd case, both terminal exon misalignment and fake exons are present")
            logger.debug(isoform_exons)
            logger.debug(read_exons)
        if event_to_remove != -1:
            del matching_events[event_to_remove]

        new_events = self.check_if_close(isoform_start, polya_info.external_polyt_pos, polya_info.internal_polyt_pos,
                                         matching_events, MatchEventSubtype.correct_polya_site_left)
        if new_events is not None:
            logger.debug("Orginal polyT seems good")
            return new_events

        logger.debug("Orginal polyT seems distant, checking for fake terminal exons")
        # if terminal exon is fake, move polyT accordingly
        external_polyt_pos, internal_polyt_pos = \
            self.correct_polyt_positions(read_exons, fake_terminal_exon_count, polya_info)
        logger.debug("+ Corrected external %d, corrected internal %d" % (external_polyt_pos, internal_polyt_pos))

        if terminal_exon_misaligned:
            # we believe isofrom end is a true polyT site
            external_polyt_pos = isoform_exons[0][0]
            internal_polyt_pos = isoform_exons[0][0]
        else:
            matching_events, external_polyt_pos, internal_polyt_pos = \
                self.detect_reference_exons_before_polyt(isoform_exons, external_polyt_pos, internal_polyt_pos, matching_events)

        new_events = self.check_if_close(isoform_start, external_polyt_pos, internal_polyt_pos,
                                         matching_events, MatchEventSubtype.correct_polya_site_left)
        if new_events is not None:
            logger.debug("Corrected polyT seems good")
            return new_events

        polyt_pos = external_polyt_pos if internal_polyt_pos == -1 else internal_polyt_pos
        dist_to_polyt = abs(polyt_pos - isoform_start)
        logger.debug("+ Distance to polyT is %d" % dist_to_polyt)
        if dist_to_polyt > self.params.apa_delta:
            logger.debug("+ Seems like APA site")
            matching_events.append(MatchEvent(MatchEventSubtype.alternative_polya_site_left, event_info=polyt_pos))
        else:
            logger.debug("+ Seems like correct polyT, odd case")
            matching_events.append(MatchEvent(MatchEventSubtype.correct_polya_site_left, event_info=polyt_pos))
        return matching_events

    # check if polyA found within intron and inside mapped part of the read
    def check_internal_polya(self, internal_polya_pos, matching_events):
        if internal_polya_pos == -1:
            return matching_events, False

        for i, event in enumerate(matching_events):
            if event.event_type == MatchEventSubtype.incomplete_intron_retention_right:
                matching_events.append(MatchEvent(MatchEventSubtype.internal_polya_right,
                                                  isoform_region=event.isoform_region,
                                                  event_info=internal_polya_pos))
                return matching_events, True

        return matching_events, False

    # check if polyT found within intron and inside mapped part of the read
    def check_internal_polyt(self, internal_polyt_pos, matching_events):
        if internal_polyt_pos == -1:
            return matching_events, False

        for i, event in enumerate(matching_events):
            if event.event_type == MatchEventSubtype.incomplete_intron_retention_left:
                # polyT found within intron and inside mapped part of the read
                matching_events.append(MatchEvent(MatchEventSubtype.internal_polya_left,
                                                  isoform_region=event.isoform_region,
                                                  event_info=internal_polyt_pos))
                return matching_events, True

        return matching_events, False

    # correct polyA position when fake terminal exons are present
    def correct_polya_positions(self, read_exons, fake_terminal_exon_count, polya_info):
        assert fake_terminal_exon_count < len(read_exons)

        return shift_polya(read_exons, fake_terminal_exon_count, polya_info.external_polya_pos), \
               shift_polya(read_exons, fake_terminal_exon_count, polya_info.internal_polya_pos)

    # correct polyT position when fake terminal exons are present or isoform has short terminal exons
    def correct_polyt_positions(self, read_exons, fake_terminal_exon_count, polya_info):
        assert fake_terminal_exon_count < len(read_exons)

        return shift_polyt(read_exons, fake_terminal_exon_count, polya_info.external_polyt_pos), \
               shift_polyt(read_exons, fake_terminal_exon_count, polya_info.internal_polyt_pos)

    # check isoform exons beyond polyA
    def detect_reference_exons_beyond_polya(self, isoform_exons, external_polya_pos, internal_polya_pos, matching_events):
        terminal_exon_count = 0
        polya_pos = internal_polya_pos if internal_polya_pos != -1 else external_polya_pos
        while terminal_exon_count < len(isoform_exons) and isoform_exons[-terminal_exon_count-1][0] >= polya_pos:
            terminal_exon_count += 1
        if terminal_exon_count == len(isoform_exons) or terminal_exon_count == 0:
            logger.debug("+ No reference short exons")
            return matching_events, external_polya_pos, internal_polya_pos

        isoform_terminal_exon_length = intervals_total_length(isoform_exons[-terminal_exon_count:])
        dist_to_external_polya = abs(isoform_exons[-terminal_exon_count - 1][1] - external_polya_pos)
        dist_to_internal_polya = abs(isoform_exons[-terminal_exon_count - 1][1] - internal_polya_pos)
        dist_to_polya = min(dist_to_external_polya, dist_to_internal_polya)

        if (isoform_terminal_exon_length <= self.params.max_fake_terminal_exon_len and
                dist_to_polya <= self.params.max_fake_terminal_exon_len) or \
                (isoform_terminal_exon_length <= self.params.max_missed_exon_len and
                 abs(isoform_terminal_exon_length - dist_to_polya) <= self.params.delta):
            for i in range(terminal_exon_count):
                isoform_region = (len(isoform_exons) - 2 - i, len(isoform_exons) - 2 - i)
                matching_events.append(MatchEvent(MatchEventSubtype.terminal_exon_misalignment_right, isoform_region))
            corrected_read_end = isoform_exons[-1][1]
            logger.debug("+ Looks like missed terminal exons, shifting polyA: %d" % corrected_read_end)
            return matching_events, corrected_read_end, corrected_read_end

        logger.debug("+ Reference short exons present but not ok")
        return matching_events, external_polya_pos, internal_polya_pos

    # check isoform exons prior polyT
    def detect_reference_exons_before_polyt(self, isoform_exons, external_polyt_pos, internal_polyt_pos, matching_events):
        # check isoform exons beyond polyA
        terminal_exon_count = 0
        polyt_pos = internal_polyt_pos if internal_polyt_pos != -1 else external_polyt_pos
        while terminal_exon_count < len(isoform_exons) and isoform_exons[terminal_exon_count][1] <= polyt_pos:
            terminal_exon_count += 1
        if terminal_exon_count == 0 or terminal_exon_count == len(isoform_exons):
            logger.debug("+ No reference short exons")
            return matching_events, external_polyt_pos, internal_polyt_pos

        isoform_terminal_exon_length = intervals_total_length(isoform_exons[:terminal_exon_count])
        dist_to_external_polyt = abs(isoform_exons[terminal_exon_count][0] - external_polyt_pos)
        dist_to_internal_polyt = abs(isoform_exons[terminal_exon_count][0] - internal_polyt_pos)
        dist_to_polyt = min(dist_to_external_polyt, dist_to_internal_polyt)
        # if it looks like we did not align last isoform exons
        if (isoform_terminal_exon_length <= self.params.max_fake_terminal_exon_len and
                dist_to_polyt <= self.params.max_fake_terminal_exon_len) or \
                (isoform_terminal_exon_length <= self.params.max_missed_exon_len and
                 abs(isoform_terminal_exon_length - dist_to_polyt) <= self.params.delta):
            for i in range(terminal_exon_count):
                matching_events.append(MatchEvent(MatchEventSubtype.terminal_exon_misalignment_left, (i, i)))
            corrected_read_start = isoform_exons[0][0]
            logger.debug("+ Looks like missed terminal exons, shifting polyT: %d" % corrected_read_start)
            return matching_events, corrected_read_start, corrected_read_start

        logger.debug("+ Reference short exons present but not ok")
        return matching_events, external_polyt_pos, internal_polyt_pos

    def check_if_close(self, isoform_end, external_polya_pos, internal_polya_pos, matching_events, event_type):
        dist_to_external_polya = abs(isoform_end - external_polya_pos) if external_polya_pos != -1 else math.inf
        dist_to_internal_polya = abs(isoform_end - internal_polya_pos) if internal_polya_pos != -1 else math.inf

        # polyA position is close, we are satisfied
        if  dist_to_internal_polya <= self.params.apa_delta and dist_to_internal_polya <= dist_to_external_polya:
            logger.debug("Internal polyAT is good %d, distance %d" %
                         (internal_polya_pos, dist_to_internal_polya))
            matching_events.append(MatchEvent(event_type, event_info=internal_polya_pos))
            return matching_events
        elif dist_to_external_polya <= self.params.apa_delta and dist_to_internal_polya > dist_to_external_polya:
            logger.debug("External polyAT is good %d, distance %d" %
                         (external_polya_pos, dist_to_external_polya))
            matching_events.append(MatchEvent(event_type, event_info=external_polya_pos))
            return matching_events
        return None



############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Tests for molecule structure parsing and element types.

Tests the universal barcode calling infrastructure:
- ElementType enum and its method classification
- MoleculeElement initialization for different element types
- MoleculeStructure parsing from MDF format
"""

import os
import pytest
from io import StringIO

from src.barcode_calling.callers.molecule_structure import (
    ElementType,
    MoleculeElement,
    MoleculeStructure
)

# Test data directory
TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "universal_data")


class TestElementType:
    """Test ElementType enum methods."""

    def test_needs_correction_var_file(self):
        """VAR_FILE elements need barcode correction."""
        assert ElementType.VAR_FILE.needs_correction() is True

    def test_needs_correction_var_list(self):
        """VAR_LIST elements need barcode correction."""
        assert ElementType.VAR_LIST.needs_correction() is True

    def test_needs_correction_var_any(self):
        """VAR_ANY elements do NOT need correction."""
        assert ElementType.VAR_ANY.needs_correction() is False

    def test_needs_correction_const(self):
        """CONST elements do NOT need correction."""
        assert ElementType.CONST.needs_correction() is False

    def test_needs_sequence_extraction_var_any(self):
        """VAR_ANY elements need sequence extraction."""
        assert ElementType.VAR_ANY.needs_sequence_extraction() is True

    def test_needs_sequence_extraction_var_file(self):
        """VAR_FILE elements do NOT need sequence extraction (they get corrected)."""
        assert ElementType.VAR_FILE.needs_sequence_extraction() is False

    def test_needs_sequence_extraction_const(self):
        """CONST elements do NOT need sequence extraction."""
        assert ElementType.CONST.needs_sequence_extraction() is False

    def test_is_constant_const(self):
        """CONST is a constant element."""
        assert ElementType.CONST.is_constant() is True

    def test_is_constant_var_any(self):
        """VAR_ANY is NOT a constant element."""
        assert ElementType.VAR_ANY.is_constant() is False

    def test_is_constant_polyt(self):
        """PolyT is NOT a constant element."""
        assert ElementType.PolyT.is_constant() is False

    def test_is_variable_var_any(self):
        """VAR_ANY is a variable element."""
        assert ElementType.VAR_ANY.is_variable() is True

    def test_is_variable_var_list(self):
        """VAR_LIST is a variable element."""
        assert ElementType.VAR_LIST.is_variable() is True

    def test_is_variable_var_file(self):
        """VAR_FILE is a variable element."""
        assert ElementType.VAR_FILE.is_variable() is True

    def test_is_variable_const(self):
        """CONST is NOT a variable element."""
        assert ElementType.CONST.is_variable() is False

    def test_is_variable_polyt(self):
        """PolyT is NOT a variable element."""
        assert ElementType.PolyT.is_variable() is False

    def test_is_base_separator_var_any_separator(self):
        """VAR_ANY_SEPARATOR is a separator."""
        assert ElementType.VAR_ANY_SEPARATOR.is_base_separator() is True

    def test_is_base_separator_var_any_non_t_separator(self):
        """VAR_ANY_NON_T_SEPARATOR is a separator."""
        assert ElementType.VAR_ANY_NON_T_SEPARATOR.is_base_separator() is True

    def test_is_base_separator_var_any(self):
        """VAR_ANY is NOT a separator."""
        assert ElementType.VAR_ANY.is_base_separator() is False


class TestMoleculeElement:
    """Test MoleculeElement initialization."""

    def test_init_polyt(self):
        """Test PolyT element initialization."""
        element = MoleculeElement("PolyT", ElementType.PolyT)

        assert element.element_name == "PolyT"
        assert element.element_type == ElementType.PolyT
        assert element.element_value is None
        assert element.element_length == -1

    def test_init_cdna(self):
        """Test cDNA element initialization."""
        element = MoleculeElement("cDNA", ElementType.cDNA)

        assert element.element_name == "cDNA"
        assert element.element_type == ElementType.cDNA
        assert element.element_value is None
        assert element.element_length == -1

    def test_init_const(self):
        """Test CONST element initialization."""
        sequence = "CTACACGACGCTCTTCCGATCT"
        element = MoleculeElement("R1", ElementType.CONST, sequence)

        assert element.element_name == "R1"
        assert element.element_type == ElementType.CONST
        assert element.element_value == sequence
        assert element.element_length == len(sequence)

    def test_init_var_list(self):
        """Test VAR_LIST element initialization."""
        barcodes = "AAAA,CCCC,GGGG,TTTT"
        element = MoleculeElement("Barcode", ElementType.VAR_LIST, barcodes)

        assert element.element_name == "Barcode"
        assert element.element_type == ElementType.VAR_LIST
        assert element.element_value == ["AAAA", "CCCC", "GGGG", "TTTT"]
        assert element.element_length == 4  # Length of first barcode

    def test_init_var_any(self):
        """Test VAR_ANY element initialization."""
        element = MoleculeElement("UMI", ElementType.VAR_ANY, "12")

        assert element.element_name == "UMI"
        assert element.element_type == ElementType.VAR_ANY
        assert element.element_value is None
        assert element.element_length == 12

    def test_init_var_file(self):
        """Test VAR_FILE element initialization with real file."""
        barcode_file = os.path.join(TEST_DATA_DIR, "barcodes_small.tsv")
        element = MoleculeElement("Barcode", ElementType.VAR_FILE, barcode_file)

        assert element.element_name == "Barcode"
        assert element.element_type == ElementType.VAR_FILE
        assert len(element.element_value) == 8  # 8 barcodes in test file
        assert element.element_length == 16  # 16bp barcodes

    def test_init_var_any_separator(self):
        """Test VAR_ANY_SEPARATOR element initialization."""
        element = MoleculeElement("Sep", ElementType.VAR_ANY_SEPARATOR, "5")

        assert element.element_name == "Sep"
        assert element.element_type == ElementType.VAR_ANY_SEPARATOR
        assert element.element_value is None
        assert element.element_length == 5


class TestMoleculeStructure:
    """Test MoleculeStructure parsing."""

    def test_parse_simple_structure(self):
        """Test parsing simple molecule structure."""
        mdf_content = """Barcode:UMI:PolyT:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC,GGGG
UMI\tVAR_ANY\t10
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        elements = list(structure)
        assert len(elements) == 4

        assert elements[0].element_name == "Barcode"
        assert elements[0].element_type == ElementType.VAR_LIST

        assert elements[1].element_name == "UMI"
        assert elements[1].element_type == ElementType.VAR_ANY

        assert elements[2].element_name == "PolyT"
        assert elements[2].element_type == ElementType.PolyT

        assert elements[3].element_name == "cDNA"
        assert elements[3].element_type == ElementType.cDNA

    def test_parse_with_const(self):
        """Test parsing structure with constant element."""
        mdf_content = """R1:Barcode:UMI:PolyT:cDNA
R1\tCONST\tCTACACGACGCTCTTCCGATCT
Barcode\tVAR_LIST\tAAAA,CCCC,GGGG
UMI\tVAR_ANY\t12
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        elements = list(structure)
        assert len(elements) == 5

        assert elements[0].element_name == "R1"
        assert elements[0].element_type == ElementType.CONST
        assert elements[0].element_value == "CTACACGACGCTCTTCCGATCT"

    def test_parse_from_file_simple(self):
        """Test parsing simple.mdf file."""
        mdf_path = os.path.join(TEST_DATA_DIR, "simple.mdf")

        with open(mdf_path) as f:
            structure = MoleculeStructure(f)

        elements = list(structure)

        # Simple.mdf: Barcode:UMI:PolyT:cDNA
        assert len(elements) == 4
        assert elements[0].element_name == "Barcode"
        assert elements[0].element_type == ElementType.VAR_LIST
        assert elements[1].element_name == "UMI"
        assert elements[1].element_type == ElementType.VAR_ANY
        assert elements[1].element_length == 10

    def test_parse_from_file_10x(self):
        """Test parsing 10x-like structure.

        Uses inline definition to avoid relative path issues with VAR_FILE.
        """
        # Use inline definition to avoid file path issues
        mdf_content = """R1:Barcode:UMI:PolyT:cDNA:TSO
R1\tCONST\tCTACACGACGCTCTTCCGATCT
Barcode\tVAR_LIST\tATCCTTAGTGTTTGTC,CTTAGGAGTGGTTAGC,TGAGGGCCAACGTGCT
UMI\tVAR_ANY\t12
TSO\tCONST\tCCCATGTACTCTGCGTTGATACCACTGCTT
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        elements = list(structure)

        # R1:Barcode:UMI:PolyT:cDNA:TSO
        assert len(elements) == 6
        assert elements[0].element_name == "R1"
        assert elements[0].element_type == ElementType.CONST
        assert elements[1].element_name == "Barcode"
        assert elements[1].element_type == ElementType.VAR_LIST

    def test_iteration(self):
        """Test iterating over structure elements."""
        mdf_content = """Barcode:UMI:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
UMI\tVAR_ANY\t8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        names = [el.element_name for el in structure]
        assert names == ["Barcode", "UMI", "cDNA"]

    def test_from_element_list(self):
        """Test creating structure from element list."""
        elements = [
            MoleculeElement("Barcode", ElementType.VAR_LIST, "AAAA,CCCC"),
            MoleculeElement("UMI", ElementType.VAR_ANY, "10"),
            MoleculeElement("cDNA", ElementType.cDNA)
        ]

        structure = MoleculeStructure.from_element_list(elements)

        assert len(list(structure)) == 3
        assert list(structure)[0].element_name == "Barcode"

    def test_concat_delimiter(self):
        """Test concatenation delimiter constant."""
        assert MoleculeStructure.CONCAT_DELIM == "|"

    def test_dupl_delimiter(self):
        """Test duplication delimiter constant."""
        assert MoleculeStructure.DUPL_DELIM == "/"


class TestMoleculeStructureBarcodeUMI:
    """Test barcode/UMI element identification."""

    def test_barcode_elements_single(self):
        """Test single barcode element identification."""
        mdf_content = """Barcode:UMI:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
UMI\tVAR_ANY\t10
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.barcode_elements == ["Barcode"]

    def test_umi_elements_single(self):
        """Test single UMI element identification."""
        mdf_content = """Barcode:UMI:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
UMI\tVAR_ANY\t10
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.umi_elements == ["UMI"]

    def test_barcode_elements_multiple(self):
        """Test multiple barcode elements identification."""
        mdf_content = """barcode1:barcode2:UMI:cDNA
barcode1\tVAR_LIST\tAAAA,CCCC
barcode2\tVAR_LIST\tGGGG,TTTT
UMI\tVAR_ANY\t10
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.barcode_elements == ["barcode1", "barcode2"]

    def test_umi_elements_multiple(self):
        """Test multiple UMI elements identification."""
        mdf_content = """Barcode:UMI1:UMI2:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
UMI1\tVAR_ANY\t6
UMI2\tVAR_ANY\t6
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.umi_elements == ["UMI1", "UMI2"]

    def test_barcode_case_insensitive(self):
        """Test that barcode prefix is case-insensitive."""
        mdf_content = """BARCODE:Barcode2:barcode3:cDNA
BARCODE\tVAR_LIST\tAAAA,CCCC
Barcode2\tVAR_LIST\tGGGG,TTTT
barcode3\tVAR_LIST\tACGT,TGCA
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert len(structure.barcode_elements) == 3
        assert "BARCODE" in structure.barcode_elements
        assert "Barcode2" in structure.barcode_elements
        assert "barcode3" in structure.barcode_elements

    def test_umi_case_insensitive(self):
        """Test that UMI prefix is case-insensitive."""
        mdf_content = """Barcode:UMI:umi2:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
UMI\tVAR_ANY\t6
umi2\tVAR_ANY\t6
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert len(structure.umi_elements) == 2
        assert "UMI" in structure.umi_elements
        assert "umi2" in structure.umi_elements

    def test_no_barcode_elements(self):
        """Test structure with no barcode elements."""
        mdf_content = """R1:UMI:cDNA
R1\tCONST\tACGT
UMI\tVAR_ANY\t10
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.barcode_elements == []

    def test_no_umi_elements(self):
        """Test structure with no UMI elements."""
        mdf_content = """Barcode:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.umi_elements == []

    def test_from_element_list_identifies_barcodes(self):
        """Test that from_element_list also identifies barcode/UMI elements."""
        elements = [
            MoleculeElement("Barcode", ElementType.VAR_LIST, "AAAA,CCCC"),
            MoleculeElement("UMI", ElementType.VAR_ANY, "10"),
            MoleculeElement("cDNA", ElementType.cDNA)
        ]

        structure = MoleculeStructure.from_element_list(elements)

        assert structure.barcode_elements == ["Barcode"]
        assert structure.umi_elements == ["UMI"]


class TestMoleculeStructureEdgeCases:
    """Test edge cases and error handling."""

    def test_element_order_preserved(self):
        """Test that element order from header line is preserved."""
        mdf_content = """cDNA:UMI:Barcode:PolyT
Barcode\tVAR_LIST\tAAAA,CCCC
UMI\tVAR_ANY\t8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        names = [el.element_name for el in structure]
        # Order should match header line, not definition order
        assert names == ["cDNA", "UMI", "Barcode", "PolyT"]

    def test_whitespace_handling(self):
        """Test that whitespace in MDF is handled correctly."""
        mdf_content = """  Barcode : UMI : cDNA
Barcode    VAR_LIST    AAAA,CCCC
UMI        VAR_ANY     8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        elements = list(structure)
        assert len(elements) == 3

    def test_multiple_var_list_barcodes(self):
        """Test VAR_LIST with many barcodes."""
        barcodes = ",".join(["ACGT" * 4] * 100)  # 100 barcodes
        mdf_content = f"""Barcode:cDNA
Barcode\tVAR_LIST\t{barcodes}
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        elements = list(structure)
        assert len(elements[0].element_value) == 100


class TestConcatenatedElements:
    """Test concatenated element parsing and identification."""

    def test_parse_concatenated_elements(self):
        """Test parsing MDF with concatenated barcode parts."""
        mdf_content = """Barcode|1:Barcode|2:UMI:PolyT:cDNA
Barcode|1\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
Barcode|2\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
UMI\tVAR_ANY\t8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert len(structure.ordered_elements) == 5
        assert "Barcode|1" in structure.elements_to_concatenate
        assert "Barcode|2" in structure.elements_to_concatenate
        assert structure.elements_to_concatenate["Barcode|1"] == ("Barcode", 1)
        assert structure.elements_to_concatenate["Barcode|2"] == ("Barcode", 2)

    def test_concatenated_elements_counts(self):
        """Test that concatenated_elements_counts has correct base name and count."""
        mdf_content = """Barcode|1:Barcode|2:cDNA
Barcode|1\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
Barcode|2\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert "Barcode" in structure.concatenated_elements_counts
        assert structure.concatenated_elements_counts["Barcode"] == 2

    def test_concatenated_three_parts(self):
        """Test three-part concatenated element."""
        mdf_content = """Barcode|1:Barcode|2:Barcode|3:cDNA
Barcode|1\tVAR_LIST\tAAAACCCCGGGG,TTTTAAAACCCC\t4
Barcode|2\tVAR_LIST\tAAAACCCCGGGG,TTTTAAAACCCC\t4
Barcode|3\tVAR_LIST\tAAAACCCCGGGG,TTTTAAAACCCC\t4
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.concatenated_elements_counts["Barcode"] == 3

    def test_concatenated_barcode_umi_identification(self):
        """Test that concatenated barcode elements are identified by base name."""
        mdf_content = """Barcode|1:Barcode|2:UMI:cDNA
Barcode|1\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
Barcode|2\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
UMI\tVAR_ANY\t8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        # Base name "Barcode" should be in barcode_elements, not "Barcode|1", "Barcode|2"
        assert structure.barcode_elements == ["Barcode"]
        assert "Barcode|1" not in structure.barcode_elements
        assert "Barcode|2" not in structure.barcode_elements
        assert structure.umi_elements == ["UMI"]

    def test_concatenated_umi_identification(self):
        """Test that concatenated UMI elements are identified by base name."""
        mdf_content = """Barcode:UMI|1:UMI|2:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
UMI|1\tVAR_ANY\t4
UMI|2\tVAR_ANY\t4
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.barcode_elements == ["Barcode"]
        assert structure.umi_elements == ["UMI"]

    def test_concatenated_explicit_length(self):
        """Test that explicit length (4th field) is correctly parsed for concatenated parts."""
        mdf_content = """Barcode|1:Barcode|2:cDNA
Barcode|1\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
Barcode|2\tVAR_LIST\tAAAACCCC,GGGGTTTT\t4
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        # Each part should have length 4, not the full barcode length 8
        for el in structure:
            if el.element_name == "Barcode|1":
                assert el.element_length == 4
            elif el.element_name == "Barcode|2":
                assert el.element_length == 4

    def test_concatenated_elements_must_be_variable(self):
        """Test that non-variable concatenated elements are rejected."""
        mdf_content = """Adapter|1:Adapter|2:cDNA
Adapter|1\tCONST\tACGT
Adapter|2\tCONST\tTGCA
"""
        with pytest.raises(SystemExit):
            MoleculeStructure(iter(mdf_content.strip().split('\n')))

    def test_concatenated_nonconsecutive_indices_rejected(self):
        """Test that non-consecutive indices (e.g., 1 and 3) are rejected."""
        mdf_content = """Barcode|1:Barcode|3:cDNA
Barcode|1\tVAR_LIST\tAAAA,CCCC\t4
Barcode|3\tVAR_LIST\tAAAA,CCCC\t4
"""
        with pytest.raises(SystemExit):
            MoleculeStructure(iter(mdf_content.strip().split('\n')))


class TestDuplicatedElements:
    """Test duplicated element parsing and identification."""

    def test_parse_duplicated_elements(self):
        """Test parsing MDF with duplicated barcode parts."""
        mdf_content = """Barcode/1:Barcode/2:UMI:PolyT:cDNA
Barcode/1\tVAR_LIST\tAAAACCCC,GGGGTTTT
Barcode/2\tVAR_LIST\tAAAACCCC,GGGGTTTT
UMI\tVAR_ANY\t8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert len(structure.ordered_elements) == 5
        assert "Barcode/1" in structure.duplicated_elements
        assert "Barcode/2" in structure.duplicated_elements
        assert structure.duplicated_elements["Barcode/1"] == ("Barcode", 1)
        assert structure.duplicated_elements["Barcode/2"] == ("Barcode", 2)

    def test_duplicated_elements_counts(self):
        """Test that duplicated_elements_counts has correct base name and count."""
        mdf_content = """Barcode/1:Barcode/2:cDNA
Barcode/1\tVAR_LIST\tAAAACCCC,GGGGTTTT
Barcode/2\tVAR_LIST\tAAAACCCC,GGGGTTTT
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert "Barcode" in structure.duplicated_elements_counts
        assert structure.duplicated_elements_counts["Barcode"] == 2

    def test_duplicated_three_parts(self):
        """Test three-part duplicated element."""
        mdf_content = """Barcode/1:Barcode/2:Barcode/3:cDNA
Barcode/1\tVAR_LIST\tAAAACCCC,GGGGTTTT
Barcode/2\tVAR_LIST\tAAAACCCC,GGGGTTTT
Barcode/3\tVAR_LIST\tAAAACCCC,GGGGTTTT
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.duplicated_elements_counts["Barcode"] == 3

    def test_duplicated_barcode_umi_identification(self):
        """Test that duplicated barcode elements are identified by base name."""
        mdf_content = """Barcode/1:Barcode/2:UMI:cDNA
Barcode/1\tVAR_LIST\tAAAACCCC,GGGGTTTT
Barcode/2\tVAR_LIST\tAAAACCCC,GGGGTTTT
UMI\tVAR_ANY\t8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        # Base name "Barcode" should be in barcode_elements, not "Barcode/1", "Barcode/2"
        assert structure.barcode_elements == ["Barcode"]
        assert "Barcode/1" not in structure.barcode_elements
        assert "Barcode/2" not in structure.barcode_elements
        assert structure.umi_elements == ["UMI"]

    def test_duplicated_umi_identification(self):
        """Test that duplicated UMI elements are identified by base name."""
        mdf_content = """Barcode:UMI/1:UMI/2:cDNA
Barcode\tVAR_LIST\tAAAA,CCCC
UMI/1\tVAR_ANY\t8
UMI/2\tVAR_ANY\t8
"""
        structure = MoleculeStructure(iter(mdf_content.strip().split('\n')))

        assert structure.barcode_elements == ["Barcode"]
        assert structure.umi_elements == ["UMI"]

    def test_duplicated_elements_must_be_variable(self):
        """Test that non-variable duplicated elements are rejected."""
        mdf_content = """Adapter/1:Adapter/2:cDNA
Adapter/1\tCONST\tACGT
Adapter/2\tCONST\tACGT
"""
        with pytest.raises(SystemExit):
            MoleculeStructure(iter(mdf_content.strip().split('\n')))

    def test_duplicated_nonconsecutive_indices_rejected(self):
        """Test that non-consecutive indices (e.g., 1 and 3) are rejected."""
        mdf_content = """Barcode/1:Barcode/3:cDNA
Barcode/1\tVAR_LIST\tAAAA,CCCC
Barcode/3\tVAR_LIST\tAAAA,CCCC
"""
        with pytest.raises(SystemExit):
            MoleculeStructure(iter(mdf_content.strip().split('\n')))

    def test_duplicated_different_types_rejected(self):
        """Test that duplicated elements with different types are rejected."""
        mdf_content = """Barcode/1:Barcode/2:cDNA
Barcode/1\tVAR_LIST\tAAAA,CCCC
Barcode/2\tVAR_ANY\t4
"""
        with pytest.raises(SystemExit):
            MoleculeStructure(iter(mdf_content.strip().split('\n')))


if __name__ == '__main__':
    pytest.main([__file__, '-v'])




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import subprocess

import pytest
import os
import shutil


def test_run_without_parameters():
    result = subprocess.run(["python3", "isoquant.py"], capture_output=True)
    assert result.returncode == 0
    assert b"usage" in result.stdout


@pytest.mark.parametrize("option", ["-h", "--help", "--full_help"])
def test__help(option):
    result = subprocess.run(["./isoquant.py", option], capture_output=True)
    print(result.returncode)
    assert result.returncode == 0
    assert b"usage" in result.stdout
    assert b"options:" in result.stdout


def test_clean_start():
    source_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(source_dir, 'simple_data/')
    out_dir = os.path.join(source_dir, "out_full/")
    shutil.rmtree(out_dir, ignore_errors=True)
    sample_name = "ONT_Simulated.chr9.4M"

    result = subprocess.run(["python3", "isoquant.py",
                             "--clean_start",
                             "-o", out_dir,
                             "--data_type", "nanopore",
                             "--fastq", data_dir + "chr9.4M.ont.sim.fq.gz",
                             "--genedb", data_dir + "chr9.4M.gtf.gz", "--complete_genedb",
                             "-r",  data_dir + "chr9.4M.fa.gz",
                             "-t", "2",
                             "--prefix", sample_name,
                             "--count_exons", "--sqanti_output",
                             "--read_group","file:" + data_dir + "chr9.4M.ont.sim.read_groups.tsv" + ":0:1"])

    assert result.returncode == 0
    sample_folder = os.path.join(out_dir, sample_name)
    assert os.path.isdir(sample_folder)
    # Note: corrected_reads.bed.gz and transcript_model_reads.tsv.gz are only generated
    # with --large_output corrected_bed read2transcripts (not in default)
    resulting_files = ["exon_counts.tsv", "exon_grouped_file0_col1_counts.linear.tsv",
                       "gene_counts.tsv", "gene_grouped_file0_col1_counts.tsv",
                       "intron_counts.tsv", "intron_grouped_file0_col1_counts.linear.tsv",
                       "read_assignments.tsv.gz",
                       "novel_vs_known.SQANTI-like.tsv",
                       "transcript_counts.tsv", "transcript_grouped_file0_col1_counts.tsv",
                       "discovered_transcript_counts.tsv", "transcript_models.gtf",
                       "discovered_transcript_tpm.tsv"]
    for f in resulting_files:
        assert os.path.exists(os.path.join(sample_folder, sample_name + "." + f))


def test_usual_start():
    source_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(source_dir, 'simple_data/')
    out_dir = os.path.join(source_dir, "out_usual/")
    shutil.rmtree(out_dir, ignore_errors=True)
    sample_name = "ONT_Simulated.chr9.4M"
    result = subprocess.run(["python3", "--version"])
    result = subprocess.run(["python3", "isoquant.py",
                             "-o", out_dir,
                             "--data_type", "nanopore",
                             "--fastq", data_dir + "chr9.4M.ont.sim.fq.gz",
                             "--genedb", data_dir + "chr9.4M.gtf.gz", "--complete_genedb",
                             "-r",  data_dir + "chr9.4M.fa.gz",
                             "-t", "2",
                             "--prefix", sample_name])

    assert result.returncode == 0
    sample_folder = os.path.join(out_dir, sample_name)
    assert os.path.isdir(sample_folder)
    resulting_files = ["gene_counts.tsv", "read_assignments.tsv.gz", "transcript_counts.tsv",
                       "discovered_transcript_counts.tsv", "transcript_models.gtf"]
    for f in resulting_files:
        assert os.path.exists(os.path.join(sample_folder, sample_name + "." + f))


def test_with_bam_and_polya():
    source_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(source_dir, 'simple_data/')
    out_dir = os.path.join(source_dir, "out_polya/")
    shutil.rmtree(out_dir, ignore_errors=True)
    sample_name = "ONT_Simulated.chr9.4M.polyA"

    result = subprocess.run(["python3", "isoquant.py",
                             "-o", out_dir,
                             "--data_type", "nanopore",
                             "--bam", os.path.join(data_dir, "chr9.4M.ont.sim.polya.bam"),
                             "--genedb", os.path.join(data_dir, "chr9.4M.gtf.gz"), "--complete_genedb",
                             "-r",  os.path.join(data_dir, "chr9.4M.fa.gz"),
                             "-t", "2",
                             "--prefix", sample_name,
                             "--sqanti_output", "--count_exons"])

    assert result.returncode == 0
    sample_folder = os.path.join(out_dir, sample_name)
    assert os.path.isdir(sample_folder)
    resulting_files = ["exon_counts.tsv", "gene_counts.tsv",
                       "intron_counts.tsv", "read_assignments.tsv.gz",
                       "novel_vs_known.SQANTI-like.tsv",
                       "transcript_counts.tsv",
                       "discovered_transcript_counts.tsv", "transcript_models.gtf"]
    for f in resulting_files:
        assert os.path.exists(os.path.join(sample_folder, sample_name + "." + f))


def test_with_illumina():
    source_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(source_dir, 'simple_data/')
    out_dir = os.path.join(source_dir, "out_illumina/")
    shutil.rmtree(out_dir, ignore_errors=True)
    sample_name = "ONT_Simulated.chr9.4M.polyA"

    result = subprocess.run(["python3", "isoquant.py",
                             "-o", out_dir,
                             "--data_type", "nanopore",
                             "--bam", os.path.join(data_dir, "chr9.4M.ont.sim.polya.bam"),
                             "--illumina_bam", os.path.join(data_dir, "chr9.4M.Illumina.bam"),
                             "--genedb", os.path.join(data_dir, "chr9.4M.gtf.gz"), "--complete_genedb",
                             "-r",  os.path.join(data_dir, "chr9.4M.fa.gz"),
                             "-t", "2",
                             "--prefix", sample_name])

    assert result.returncode == 0
    sample_folder = os.path.join(out_dir, sample_name)
    assert os.path.isdir(sample_folder)
    resulting_files = ["gene_counts.tsv",
                       "read_assignments.tsv.gz",
                       "transcript_counts.tsv",
                       "discovered_transcript_counts.tsv", "transcript_models.gtf"]
    for f in resulting_files:
        assert os.path.exists(os.path.join(sample_folder, sample_name + "." + f))

def test_with_yaml():
    source_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(source_dir, 'simple_data/')
    out_dir = os.path.join(source_dir, "out_yaml/")
    shutil.rmtree(out_dir, ignore_errors=True)
    sample_name = "ONT_Simulated.chr9.4M.polyA"

    result = subprocess.run(["python3", "isoquant.py",
                             "-o", out_dir,
                             "--data_type", "nanopore",
                             "--yaml", os.path.join(data_dir, "chr9.4M.yaml"),
                             "--genedb", os.path.join(data_dir, "chr9.4M.gtf.gz"), "--complete_genedb",
                             "-r",  os.path.join(data_dir, "chr9.4M.fa.gz"),
                             "-t", "2"])

    assert result.returncode == 0
    sample_folder = os.path.join(out_dir, sample_name)
    assert os.path.isdir(sample_folder)
    resulting_files = ["gene_counts.tsv",
                       "read_assignments.tsv.gz",
                       "transcript_counts.tsv",
                       "discovered_transcript_counts.tsv", "transcript_models.gtf"]
    for f in resulting_files:
        assert os.path.exists(os.path.join(sample_folder, sample_name + "." + f))


#def test_cage():
#    source_dir = os.path.dirname(os.path.realpath(__file__))
#    data_dir = os.path.join(source_dir, 'toy_data/')
#    out_dir = os.path.join(source_dir, "out_mapt/")
#    shutil.rmtree(out_dir, ignore_errors=True)
#    os.environ['HOME'] = source_dir # dirty hack to set $HOME for tox environment
#    sample_name = "MAPT.Mouse.ONT"
#
#    result = subprocess.run(["python", "isoquant.py", '--clean_start',
#                             "-o", out_dir,
#                             "--data_type", "nanopore",
#                             "--fastq", data_dir + "MAPT.Mouse.ONT.simulated.fastq",
#                             "--genedb", data_dir + "MAPT.Mouse.genedb.gtf", "--complete_genedb",
#                             "-r",  data_dir + "MAPT.Mouse.reference.fasta",
#                             '--cage', data_dir + "MAPT.Mouse.CAGE.bed",
#                             "-t", "2",
#                             "--prefix", sample_name])
#
    # assert result.returncode == 0
    # sample_folder = os.path.join(out_dir, sample_name)
    # assert os.path.isdir(sample_folder)
    # resulting_files = ["gene_counts.tsv", "read_assignments.tsv.gz", "transcript_counts.tsv",
    #                    "transcript_model_counts.tsv", "transcript_models.gtf", "transcript_model_reads.tsv.gz"]
    # for f in resulting_files:
    #     assert os.path.exists(os.path.join(sample_folder, sample_name + "." + f))




############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
from src.barcode_calling.indexers import (
    SharedMemoryArray2BitKmerIndexer,
    SharedMemoryIndexInfo,
    Array2BitKmerIndexer,
)
from src.barcode_calling.common import str_to_2bit


class TestSharedMemoryArray2BitKmerIndexer:
    """Test shared memory k-mer indexer."""

    def test_init(self):
        """Test basic initialization."""
        barcodes = ["ACTGACTGACTGACTGACTGACTGA"]  # 25-mer
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        assert indexer.total_sequences == 1
        assert indexer.k == 10
        assert not indexer.empty()

    def test_empty(self):
        """Test empty index."""
        indexer = SharedMemoryArray2BitKmerIndexer([], kmer_size=10, seq_len=25)
        assert indexer.empty()
        assert indexer.total_sequences == 0
        assert indexer.index_size == 0

    def test_exact_match(self):
        """Test finding exact match."""
        seq = "ACTGACTGACTGACTGACTGACTGA"
        bin_seqs = [str_to_2bit(seq)]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        results = indexer.get_occurrences(seq)
        assert len(results) > 0
        assert results[0][0] == seq

    def test_similar_match(self):
        """Test finding similar sequence (1 mismatch)."""
        barcode = "ACTGACTGACTGACTGACTGACTGA"
        query = "ACTGACTGACTGACTGACTGACTGT"  # 1 mismatch at end
        bin_seqs = [str_to_2bit(barcode)]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        results = indexer.get_occurrences(query)
        assert len(results) > 0
        assert results[0][0] == barcode

    def test_consistency_with_array2bit(self):
        """Test results match Array2BitKmerIndexer."""
        barcodes = [
            "ACTGACTGACTGACTGACTGACTGA",
            "TGCATGCATGCATGCATGCATGCAT",
            "GGGGGGGGGGGGGGGGGGGGGGGGG"
        ]
        bin_seqs = [str_to_2bit(b) for b in barcodes]

        shared = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)
        regular = Array2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        query = "ACTGACTGACTGACTGACTGACTGT"
        shared_results = shared.get_occurrences(query)
        regular_results = regular.get_occurrences(query)

        # Should find same sequences
        shared_seqs = set(seq for seq, _, _ in shared_results)
        regular_seqs = set(seq for seq, _, _ in regular_results)
        assert shared_seqs == regular_seqs

    def test_pickling_sharable_info(self):
        """Test serialization via get_sharable_info/from_sharable_info."""
        barcodes = ["ACTGACTGACTGACTGACTGACTGA"]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        # Get sharable info (simulates pickling for multiprocessing)
        info = indexer.get_sharable_info()

        assert isinstance(info, SharedMemoryIndexInfo)
        assert info.barcode_count == 1
        assert info.kmer_size == 10
        assert info.seq_len == 25

        # Reconstruct from info (simulates worker process)
        reconstructed = SharedMemoryArray2BitKmerIndexer.from_sharable_info(info)

        # Both should return same results
        query = "ACTGACTGACTGACTGACTGACTGA"
        orig_results = indexer.get_occurrences(query)
        recon_results = reconstructed.get_occurrences(query)

        assert orig_results == recon_results

    def test_multiple_barcodes(self):
        """Test with multiple similar barcodes."""
        barcodes = [
            "ACTGACTGACTGACTGACTGACTGA",
            "ACTGACTGACTGACTGACTGACTGC",
            "ACTGACTGACTGACTGACTGACTGG",
        ]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        results = indexer.get_occurrences("ACTGACTGACTGACTGACTGACTGA", max_hits=2)
        assert len(results) <= 2

    def test_index_size_calculation(self):
        """Test that index_size matches actual entries."""
        barcodes = ["ACTGACTGACTGACTGACTGACTGA", "TGCATGCATGCATGCATGCATGCAT"]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        # Each barcode generates 16 k-mers (25 - 10 + 1)
        expected_entries = 2 * 16
        assert indexer.index_size == expected_entries

    def test_max_hits(self):
        """Test maximum hits limit."""
        barcodes = [
            "ACTGACTGACTGACTGACTGACTGA",
            "ACTGACTGACTGACTGACTGACTGC",
            "ACTGACTGACTGACTGACTGACTGG",
            "ACTGACTGACTGACTGACTGACTGT",
        ]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        results = indexer.get_occurrences("ACTGACTGACTGACTGACTGACTGA", max_hits=2)
        assert len(results) <= 2

    def test_min_kmers_filter(self):
        """Test minimum k-mer threshold."""
        barcodes = ["ACTGACTGACTGACTGACTGACTGA", "TGCATGCATGCATGCATGCATGCAT"]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        # Very different sequence - should not match with high min_kmers
        results = indexer.get_occurrences("CCCCCCCCCCCCCCCCCCCCCCCCC", min_kmers=10)
        assert len(results) == 0

    def test_ignore_equal(self):
        """Test ignoring exact matches."""
        barcodes = ["ACTGACTGACTGACTGACTGACTGA", "TGCATGCATGCATGCATGCATGCAT"]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = SharedMemoryArray2BitKmerIndexer(bin_seqs, kmer_size=10, seq_len=25)

        query = "ACTGACTGACTGACTGACTGACTGA"
        results = indexer.get_occurrences(query, ignore_equal=True)

        # Should not return the exact match
        for seq, _, _ in results:
            assert seq != query


class TestSharedMemoryIndexInfo:
    """Test SharedMemoryIndexInfo serialization."""

    def test_getstate_setstate(self):
        """Test pickling of SharedMemoryIndexInfo."""
        info = SharedMemoryIndexInfo(
            barcode_count=100,
            kmer_size=10,
            seq_len=25,
            index_size=1600,
            barcodes_sm_name="test_barcodes",
            index_sm_name="test_index",
            index_range_sm_name="test_ranges"
        )

        # Simulate pickle
        state = info.__getstate__()
        assert isinstance(state, tuple)
        assert len(state) == 7

        # Simulate unpickle
        new_info = SharedMemoryIndexInfo.__new__(SharedMemoryIndexInfo)
        new_info.__setstate__(state)

        assert new_info.barcode_count == 100
        assert new_info.kmer_size == 10
        assert new_info.seq_len == 25
        assert new_info.index_size == 1600
        assert new_info.barcodes_sm_name == "test_barcodes"
        assert new_info.index_sm_name == "test_index"
        assert new_info.index_range_sm_name == "test_ranges"


if __name__ == '__main__':
    pytest.main([__file__, '-v'])




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pysam
import logging

from .common import overlaps, junctions_from_blocks, get_exons
from .transcript_printer import  validate_exons

logger = logging.getLogger('IsoQuant')


class VoidExonCorrector:

    def __init__(self):
        pass

    def correct_read(self, alignment_info):
        return alignment_info.read_exons


class IlluminaExonCorrector:
    
    MAX_SCORE = 1000000000000
    ABSENT_INTRON = (0,0)
    EXON_LENGTH = 50
    SIDE_DIFF = 25
    
    def __init__(self, chromosome, start, end, short_read_file):
        self.chromosome = chromosome
        self.start = start
        self.end = end
        
        self.short_introns, self.counts = self.get_introns(short_read_file, chromosome, start, end)
        
    @classmethod
    def from_data(cls, short_introns):
        corrector = cls.__new__(cls)
        
        corrector.chromosome = 0
        corrector.start = 0
        corrector.end = 0
        corrector.short_introns = short_introns
        corrector.counts = dict()
        return corrector
        
    # get introns from a specific area of the genome
    def get_introns(self, files, chromosome, start, end):
        introns = dict()
        for f in files:
            samfile = pysam.AlignmentFile(f, "rb") 
            intr = samfile.find_introns(samfile.fetch(chromosome, start = start, stop = end))
            introns = self.merge_dictionaries(introns, intr)
            samfile.close()
            # in some cases counts might be necessary, so original is also saved
        i_list = set()
        for i in introns.keys():
            if(type(i)!="int"): 
                # gtf files start with 1, bam with 0, we use gtf as standard
                i_list.add((i[0]+1,i[1]))
        return i_list, introns
    
    @staticmethod
    def merge_dictionaries(old, new):
        old_entries = old.keys()
        for entry in new.keys():
            if entry in old_entries:
                count = new[entry] + old[entry]
                old[entry] = count
            else:
                old[entry] = new[entry]
        return old
                
    def correct_read(self, alignment_info):
        return self.correct_exons(alignment_info.read_exons)
      
    @staticmethod
    def skipped_score(left, right, old):
        return (old[0] - left[0]) + (right[1] - old[1]) - (right[0] - left[1])
        
    @staticmethod
    def better_skipped(left, right, old, score):
        return IlluminaExonCorrector.skipped_score(left, right, old) < score
        
    @staticmethod
    def right_length(left, right, old):
        middle = abs(right[0] - left[1]) <= IlluminaExonCorrector.EXON_LENGTH
        left_side = abs(old[0] - left[0]) <= IlluminaExonCorrector.SIDE_DIFF
        right_side = abs(right[1] - old[1]) <= IlluminaExonCorrector.SIDE_DIFF
        return (middle and left_side and right_side)
        
    #only one side of the new introns is allowed to be the same as the old intron
    #when both the left and the right end are the same it is often a false positive
    @staticmethod
    def one_differs(left, right, old):
        return (not left[0] == old[0] or not right[1] == old[1])

    def correct_exons(self, exons):
        introns = junctions_from_blocks(exons)
        corrected_introns = []
        score = IlluminaExonCorrector.MAX_SCORE
        sh = IlluminaExonCorrector.ABSENT_INTRON
        overlapping = []
        appended = False
        for i in introns:
            for s in self.short_introns:
                x = abs(i[0] - s[0]) + abs(i[1] - s[1])
                if overlaps(i, s):
                    #collect all overlapping introns to look for skipped exons
                    #still test if it is the best single match
                    overlapping.append(s)
                    if x < score:
                        score = x
                        sh = s
            #if (i[0] == sh[0] or i[1] == sh[1]) and sh[0] >= exons[0][0] and sh[1] <= exons[-1][1] and self.counts[(sh[0]-1,sh[1])] > 100:
            
            # if the best single match differs by 4 it is usually good to correct
            if ((i[0] == sh[0] and i[1] == sh[1]-4) or (i[1] == sh[1] and sh[0] == i[0]-4)):
            #if ((i[1] == sh[1]-4) or (sh[0] == i[0]-4)) and sh[0] >= exons[0][0] and sh[1] <= exons[-1][1]:
                corrected_introns.append(sh)
                appended = True
            #if nothing has been corrected yet check if there might be a skipped exon
            if len(overlapping) > 1 and not appended:
                score = IlluminaExonCorrector.MAX_SCORE
                left = IlluminaExonCorrector.ABSENT_INTRON
                right = IlluminaExonCorrector.ABSENT_INTRON
                for k in range(0, len(overlapping) -1):
                    x = overlapping[k]
                    for l in range(k, len(overlapping)):
                        y = overlapping[l]
                        # it is important that the exons are in the right order
                        # so there is two cases that might be a skipped exon
                        # It can only be a skipped exon if the gap between the introns has the right length
                        # as well as the differences between the old exon and the new exons
                        # additionally there might be multiple options so it also has to have the best score
                        if x[1] < y[0]: 
                            if self.right_length(x, y, i) and self.one_differs(x, y, i):
                                if self.better_skipped(x, y, i, score):
                                    left = x
                                    right = y
                                    score = self.skipped_score(x, y, i)
                        elif x[0] > y[1]:
                            if self.right_length(y, x, i) and self.one_differs(y, x, i):
                                if self.better_skipped(y, x, i, score):
                                    left = y
                                    right = x
                                    score = self.skipped_score(y, x, i)
                # if the left intron has been changed the requirements have been fulfilled and a correction can happen
                if not left == IlluminaExonCorrector.ABSENT_INTRON:
                    corrected_introns.append(left)
                    corrected_introns.append(right)
                    appended = True
            # if neither of the options lead to a correction the original intron is kept
            if not appended:
               corrected_introns.append(i)
            sh = IlluminaExonCorrector.ABSENT_INTRON
            score = IlluminaExonCorrector.MAX_SCORE
            overlapping = []
            appended = False
        # in case of introns in the wrong order the next steps will not work so debug option here
        if not validate_exons(get_exons((exons[0][0], exons[-1][1]), corrected_introns)):
            logger.debug("old:", introns)
            logger.debug("new:", corrected_introns)
        return get_exons((exons[0][0], exons[-1][1]), corrected_introns)




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pysam
import gffutils

from .common import junctions_from_blocks


def get_region_from_db(db, region): 
	
	gene_list = list(db.features_of_type('gene', limit = region, order_by=('seqid', 'start')))
	
	all_isoforms_introns = {}
	all_isoforms_exons = {}
	
	for gene_db in gene_list:
		for t in db.children(gene_db, featuretype=('transcript', 'mRNA'), order_by='start'):
			all_isoforms_exons[t.id] = []
			for e in db.children(t, order_by='start'):
				if e.featuretype == 'exon':
					all_isoforms_exons[t.id].append((e.start, e.end))

			all_isoforms_introns[t.id] = junctions_from_blocks(all_isoforms_exons[t.id])
	
	introns = set()
	for i in all_isoforms_introns.keys():
		introns.update(all_isoforms_introns[i])
		
	return introns
